from typing import Literal

AIProvider = Literal["OpenAI", "Gemini", "xAI", "UiForm"]  # , "Anthropic", "xAI"]
OpenAICompatibleProvider = Literal["OpenAI", "xAI"]  # , "xAI"]
GeminiModel = Literal[
    "gemini-2.5-pro-preview-05-06",
    "gemini-2.5-flash-preview-05-20",
    "gemini-2.5-flash-preview-04-17",
    "gemini-2.5-pro-exp-03-25",
    "gemini-2.0-flash-lite",
    "gemini-2.0-flash",
]

AnthropicModel = Literal[
    "claude-3-5-sonnet-latest", "claude-3-5-sonnet-20241022", "claude-3-5-haiku-20241022", "claude-3-opus-20240229", "claude-3-sonnet-20240229", "claude-3-haiku-20240307"
]
OpenAIModel = Literal[
    "gpt-4o",
    "gpt-4o-mini",
    "chatgpt-4o-latest",
    "gpt-4.1",
    "gpt-4.1-mini",
    "gpt-4.1-nano",
    "gpt-4o-2024-11-20",
    "gpt-4o-2024-08-06",
    "gpt-4o-2024-05-13",
    "gpt-4o-mini-2024-07-18",
    "o3-mini",
    "o3-mini-2025-01-31",
    "o1",
    "o1-2024-12-17",
    "o1-preview-2024-09-12",
    "o1-mini",
    "o1-mini-2024-09-12",
    "o3",
    "o3-2025-04-16",
    "o4-mini",
    "o4-mini-2025-04-16",
    "gpt-4.5-preview",
    "gpt-4.5-preview-2025-02-27",
    "gpt-4o-audio-preview-2024-12-17",
    "gpt-4o-audio-preview-2024-10-01",
    "gpt-4o-realtime-preview-2024-12-17",
    "gpt-4o-realtime-preview-2024-10-01",
    "gpt-4o-mini-audio-preview-2024-12-17",
    "gpt-4o-mini-realtime-preview-2024-12-17",
]
xAI_Model = Literal["grok-3-beta", "grok-3-mini-beta"]
UiFormModel = Literal["auto", "auto-small"]
LLMModel = Literal[OpenAIModel, "human", AnthropicModel, xAI_Model, GeminiModel, UiFormModel]

import datetime

from pydantic import BaseModel, Field

from uiform.types.jobs.base import InferenceSettings


class FinetunedModel(BaseModel):
    object: Literal["finetuned_model"] = "finetuned_model"
    organization_id: str
    model: str
    schema_id: str
    schema_data_id: str
    finetuning_props: InferenceSettings
    evaluation_id: str | None = None
    created_at: datetime.datetime = Field(default_factory=lambda: datetime.datetime.now(datetime.timezone.utc))


from typing import Dict, List, Literal, Optional

from pydantic import BaseModel


# Monthly Usage
class MonthlyUsageResponseContent(BaseModel):
    request_count: int


MonthlyUsageResponse = MonthlyUsageResponseContent


class Amount(BaseModel):
    value: float
    currency: str


class TokenPrice(BaseModel):
    """
    Holds pricing information (price per 1M tokens) for one token category.
    (For example, for text tokens used in the prompt.)
    """

    prompt: float  # Price per 1M prompt tokens.
    completion: float  # Price per 1M completion tokens.
    cached_discount: float = 0.5  # Price discount for cached tokens.


class Pricing(BaseModel):
    """
    Contains all pricing information that is useful for a given model.
    (For example, the grid below shows that for gpt-4o-2024-08-06 text prompt tokens
     cost $2.50 per 1M tokens while completion tokens cost $10.00 per 1M tokens.)
    """

    text: TokenPrice
    audio: Optional[TokenPrice] = None  # May be None if the model does not support audio tokens.
    ft_price_hike: float = 1.0  # Price hike for fine-tuned models.


# Fix the type definition - use proper Python syntax
ModelModality = Literal["text", "audio", "image"]

# Define supported endpoints
EndpointType = Literal[
    "chat_completions",
    "responses",
    "assistants",
    "batch",
    "fine_tuning",
    "embeddings",
    "speech_generation",
    "translation",
    "completions_legacy",
    "image_generation",
    "transcription",
    "moderation",
    "realtime",
]

# Define supported features
FeatureType = Literal["streaming", "function_calling", "structured_outputs", "distillation", "fine_tuning", "predicted_outputs"]


class ModelCapabilities(BaseModel):
    """
    Represents the capabilities of an AI model, including supported modalities,
    endpoints, and features.
    """

    modalities: List[ModelModality]
    endpoints: List[EndpointType]
    features: List[FeatureType]


class ModelCard(BaseModel):
    """
    Model card that includes pricing and capabilities.
    """

    model: LLMModel | str  # Can be a random string for finetuned models
    pricing: Pricing
    capabilities: ModelCapabilities
    temperature_support: bool = True
    reasoning_effort_support: bool = False


# List of model cards with pricing and capabilities
openai_model_cards = [
    ########################
    ########################
    # ----------------------
    # Reasoning models
    # ----------------------
    ########################
    ########################
    # ----------------------
    # o1 family
    # ----------------------
    ModelCard(
        model="o1",
        pricing=Pricing(text=TokenPrice(prompt=15.00, cached_discount=0.5, completion=60.00), audio=None),
        capabilities=ModelCapabilities(
            modalities=["text", "image"], endpoints=["chat_completions", "responses", "assistants", "batch"], features=["streaming", "function_calling", "structured_outputs"]
        ),
        temperature_support=False,
        reasoning_effort_support=True,
    ),
    ModelCard(
        model="o1-2024-12-17",
        pricing=Pricing(text=TokenPrice(prompt=15.00, cached_discount=0.5, completion=60.00), audio=None),
        capabilities=ModelCapabilities(
            modalities=["text", "image"], endpoints=["chat_completions", "responses", "assistants", "batch"], features=["streaming", "function_calling", "structured_outputs"]
        ),
        temperature_support=False,
        reasoning_effort_support=True,
    ),
    ModelCard(
        model="o1-preview-2024-09-12",
        pricing=Pricing(text=TokenPrice(prompt=15.00, cached_discount=0.5, completion=60.00), audio=None),
        capabilities=ModelCapabilities(modalities=["text"], endpoints=["chat_completions", "responses", "assistants", "batch"], features=["streaming", "function_calling"]),
        temperature_support=False,
        reasoning_effort_support=True,
    ),
    # ----------------------
    # o1-mini family
    # ----------------------
    ModelCard(
        model="o1-mini",
        pricing=Pricing(text=TokenPrice(prompt=1.10, cached_discount=0.5, completion=4.40), audio=None),
        capabilities=ModelCapabilities(modalities=["text"], endpoints=["chat_completions", "responses", "assistants"], features=["streaming"]),
        temperature_support=False,
        reasoning_effort_support=True,
    ),
    ModelCard(
        model="o1-mini-2024-09-12",
        pricing=Pricing(text=TokenPrice(prompt=1.10, cached_discount=0.5, completion=4.40), audio=None),
        capabilities=ModelCapabilities(modalities=["text"], endpoints=["chat_completions", "responses", "assistants"], features=["streaming"]),
        temperature_support=False,
        reasoning_effort_support=True,
    ),
    # ----------------------
    # o3 family
    # ----------------------
    ModelCard(
        model="o3",
        pricing=Pricing(text=TokenPrice(prompt=10.0, cached_discount=2.5 / 10.0, completion=40.0), audio=None, ft_price_hike=1.5),
        capabilities=ModelCapabilities(
            modalities=["text", "image"], endpoints=["chat_completions", "responses", "assistants", "batch"], features=["streaming", "function_calling", "structured_outputs"]
        ),
        temperature_support=False,
        reasoning_effort_support=True,
    ),
    ModelCard(
        model="o3-2025-04-16",
        pricing=Pricing(text=TokenPrice(prompt=10.0, cached_discount=2.5 / 10.0, completion=40.0), audio=None, ft_price_hike=1.5),
        capabilities=ModelCapabilities(
            modalities=["text", "image"], endpoints=["chat_completions", "responses", "assistants", "batch"], features=["streaming", "function_calling", "structured_outputs"]
        ),
        temperature_support=False,
        reasoning_effort_support=True,
    ),
    # ----------------------
    # o3-mini family
    # ----------------------
    ModelCard(
        model="o3-mini-2025-01-31",
        pricing=Pricing(text=TokenPrice(prompt=1.10, cached_discount=0.5, completion=4.40), audio=None, ft_price_hike=1.5),
        capabilities=ModelCapabilities(
            modalities=["text"], endpoints=["chat_completions", "responses", "assistants", "batch"], features=["streaming", "function_calling", "structured_outputs"]
        ),
        temperature_support=False,
        reasoning_effort_support=True,
    ),
    ModelCard(
        model="o3-mini",
        pricing=Pricing(text=TokenPrice(prompt=1.10, cached_discount=0.5, completion=4.40), audio=None, ft_price_hike=1.5),
        capabilities=ModelCapabilities(
            modalities=["text"], endpoints=["chat_completions", "responses", "assistants", "batch"], features=["streaming", "function_calling", "structured_outputs"]
        ),
        temperature_support=False,
        reasoning_effort_support=True,
    ),
    # ----------------------
    # o4-mini family
    # ----------------------
    ModelCard(
        model="o4-mini",
        pricing=Pricing(text=TokenPrice(prompt=1.10, cached_discount=0.275 / 1.1, completion=4.40), audio=None, ft_price_hike=1.5),
        capabilities=ModelCapabilities(
            modalities=["text", "image"], endpoints=["chat_completions", "responses", "assistants", "batch"], features=["streaming", "function_calling", "structured_outputs"]
        ),
        temperature_support=False,
        reasoning_effort_support=True,
    ),
    ModelCard(
        model="o4-mini-2025-04-16",
        pricing=Pricing(text=TokenPrice(prompt=1.10, cached_discount=0.275 / 1.1, completion=4.40), audio=None, ft_price_hike=1.5),
        capabilities=ModelCapabilities(
            modalities=["text", "image"], endpoints=["chat_completions", "responses", "assistants", "batch"], features=["streaming", "function_calling", "structured_outputs"]
        ),
        temperature_support=False,
        reasoning_effort_support=True,
    ),
    ########################
    ########################
    # ----------------------
    # Chat models
    # ----------------------
    ########################
    ########################
    # ----------------------
    # gpt-4.1 family
    # ----------------------
    ModelCard(
        model="gpt-4.1",
        pricing=Pricing(text=TokenPrice(prompt=2.00, cached_discount=0.25, completion=8.00), audio=None, ft_price_hike=1.5),
        capabilities=ModelCapabilities(
            modalities=["text", "image"], endpoints=["chat_completions", "responses", "assistants", "batch"], features=["streaming", "function_calling", "structured_outputs"]
        ),
        temperature_support=True,
        reasoning_effort_support=False,
    ),
    ModelCard(
        model="gpt-4.1-2025-04-14",
        pricing=Pricing(text=TokenPrice(prompt=2.00, cached_discount=0.25, completion=8.00), audio=None, ft_price_hike=1.5),
        capabilities=ModelCapabilities(
            modalities=["text", "image"], endpoints=["chat_completions", "responses", "assistants", "batch"], features=["streaming", "function_calling", "structured_outputs"]
        ),
        temperature_support=True,
        reasoning_effort_support=False,
    ),
    ModelCard(
        model="gpt-4.1-mini",
        pricing=Pricing(text=TokenPrice(prompt=0.40, cached_discount=0.25, completion=1.60), audio=None, ft_price_hike=1.5),
        capabilities=ModelCapabilities(
            modalities=["text", "image"], endpoints=["chat_completions", "responses", "assistants", "batch"], features=["streaming", "function_calling", "structured_outputs"]
        ),
        temperature_support=True,
        reasoning_effort_support=False,
    ),
    ModelCard(
        model="gpt-4.1-mini-2025-04-14",
        pricing=Pricing(text=TokenPrice(prompt=0.40, cached_discount=0.25, completion=1.60), audio=None, ft_price_hike=1.5),
        capabilities=ModelCapabilities(
            modalities=["text", "image"], endpoints=["chat_completions", "responses", "assistants", "batch"], features=["streaming", "function_calling", "structured_outputs"]
        ),
        temperature_support=True,
        reasoning_effort_support=False,
    ),
    ModelCard(
        model="gpt-4.1-nano",
        pricing=Pricing(text=TokenPrice(prompt=0.10, cached_discount=0.25, completion=0.40), audio=None, ft_price_hike=1.5),
        capabilities=ModelCapabilities(
            modalities=["text", "image"], endpoints=["chat_completions", "responses", "assistants", "batch"], features=["streaming", "function_calling", "structured_outputs"]
        ),
        temperature_support=True,
        reasoning_effort_support=False,
    ),
    ModelCard(
        model="gpt-4.1-nano-2025-04-14",
        pricing=Pricing(text=TokenPrice(prompt=0.10, cached_discount=0.25, completion=0.40), audio=None, ft_price_hike=1.5),
        capabilities=ModelCapabilities(
            modalities=["text", "image"], endpoints=["chat_completions", "responses", "assistants", "batch"], features=["streaming", "function_calling", "structured_outputs"]
        ),
        temperature_support=True,
        reasoning_effort_support=False,
    ),
    # ----------------------
    # gpt-4.5 family
    # ----------------------
    # ModelCard(
    #     model="gpt-4.5-preview",
    #     pricing=Pricing(
    #         text=TokenPrice(prompt=75, cached_discount=0.5, completion=150.00),
    #         audio=None,
    #         ft_price_hike=1.5
    #     ),
    #     capabilities=ModelCapabilities(
    #         modalities=["text", "image"],
    #         endpoints=["chat_completions", "responses", "assistants", "batch"],
    #         features=["streaming", "function_calling", "structured_outputs"]
    #     ),
    #     logprobs_support=False,
    # ),
    # ModelCard(
    #     model="gpt-4.5-preview-2025-02-27",
    #     pricing=Pricing(
    #         text=TokenPrice(prompt=75, cached_discount=0.5, completion=150.00),
    #         audio=None,
    #         ft_price_hike=1.5
    #     ),
    #     capabilities=ModelCapabilities(
    #         modalities=["text", "image"],
    #         endpoints=["chat_completions", "responses", "assistants", "batch"],
    #         features=["streaming", "function_calling", "structured_outputs"]
    #     ),
    #     logprobs_support=False,
    # ),
    # ----------------------
    # gpt-4o family
    # ----------------------
    ModelCard(
        model="chatgpt-4o-latest",
        pricing=Pricing(text=TokenPrice(prompt=2.50, cached_discount=0.5, completion=10.00), audio=None, ft_price_hike=1.5),
        capabilities=ModelCapabilities(
            modalities=["text", "image"],
            endpoints=["chat_completions", "responses", "assistants", "batch", "fine_tuning"],
            features=["streaming", "function_calling", "structured_outputs", "fine_tuning", "distillation", "predicted_outputs"],
        ),
    ),
    ModelCard(
        model="gpt-4o",
        pricing=Pricing(text=TokenPrice(prompt=2.50, cached_discount=0.5, completion=10.00), audio=None, ft_price_hike=1.5),
        capabilities=ModelCapabilities(
            modalities=["text", "image"],
            endpoints=["chat_completions", "responses", "assistants", "batch", "fine_tuning"],
            features=["streaming", "function_calling", "structured_outputs", "fine_tuning", "distillation", "predicted_outputs"],
        ),
    ),
    ModelCard(
        model="gpt-4o-2024-08-06",
        pricing=Pricing(text=TokenPrice(prompt=2.50, cached_discount=0.5, completion=10.00), audio=None, ft_price_hike=1.5),
        capabilities=ModelCapabilities(
            modalities=["text", "image"],
            endpoints=["chat_completions", "responses", "assistants", "batch", "fine_tuning"],
            features=["streaming", "function_calling", "structured_outputs", "fine_tuning", "distillation", "predicted_outputs"],
        ),
    ),
    ModelCard(
        model="gpt-4o-2024-11-20",
        pricing=Pricing(text=TokenPrice(prompt=2.50, cached_discount=0.5, completion=10.00), audio=None, ft_price_hike=1.5),
        capabilities=ModelCapabilities(
            modalities=["text", "image"],
            endpoints=["chat_completions", "responses", "assistants", "batch", "fine_tuning"],
            features=["streaming", "function_calling", "structured_outputs", "fine_tuning", "distillation", "predicted_outputs"],
        ),
    ),
    ModelCard(
        model="gpt-4o-2024-05-13",
        pricing=Pricing(text=TokenPrice(prompt=5.00, cached_discount=0.5, completion=15.00), audio=None, ft_price_hike=1.5),
        capabilities=ModelCapabilities(
            modalities=["text", "image"],
            endpoints=["chat_completions", "responses", "assistants", "batch", "fine_tuning"],
            features=["streaming", "function_calling", "structured_outputs", "fine_tuning", "distillation", "predicted_outputs"],
        ),
    ),
    # ----------------------
    # gpt-4o-audio family
    # ----------------------
    ModelCard(
        model="gpt-4o-audio-preview-2024-12-17",
        pricing=Pricing(text=TokenPrice(prompt=2.50, cached_discount=0.5, completion=10.00), audio=TokenPrice(prompt=40.00, cached_discount=0.2, completion=80.00)),
        capabilities=ModelCapabilities(modalities=["text", "audio"], endpoints=["chat_completions"], features=["streaming", "function_calling"]),
    ),
    ModelCard(
        model="gpt-4o-audio-preview-2024-10-01",
        pricing=Pricing(text=TokenPrice(prompt=2.50, cached_discount=0.5, completion=10.00), audio=TokenPrice(prompt=100.00, cached_discount=0.2, completion=200.00)),
        capabilities=ModelCapabilities(modalities=["text", "audio"], endpoints=["chat_completions"], features=["streaming", "function_calling"]),
    ),
    ModelCard(
        model="gpt-4o-realtime-preview-2024-12-17",
        pricing=Pricing(text=TokenPrice(prompt=5.00, cached_discount=0.5, completion=20.00), audio=TokenPrice(prompt=40.00, cached_discount=0.2, completion=80.00)),
        capabilities=ModelCapabilities(modalities=["text", "audio"], endpoints=["chat_completions"], features=["streaming", "function_calling"]),
    ),
    ModelCard(
        model="gpt-4o-realtime-preview-2024-10-01",
        pricing=Pricing(text=TokenPrice(prompt=5.00, cached_discount=0.5, completion=20.00), audio=TokenPrice(prompt=100.00, cached_discount=0.2, completion=200.00)),
        capabilities=ModelCapabilities(modalities=["text", "audio"], endpoints=["chat_completions"], features=["streaming", "function_calling"]),
    ),
    # ----------------------
    # gpt-4o-mini family
    # ----------------------
    ModelCard(
        model="gpt-4o-mini",
        pricing=Pricing(text=TokenPrice(prompt=0.15, cached_discount=0.5, completion=0.60), audio=None, ft_price_hike=2.0),
        capabilities=ModelCapabilities(
            modalities=["text", "image"],
            endpoints=["chat_completions", "responses", "assistants", "batch", "fine_tuning"],
            features=["streaming", "function_calling", "structured_outputs", "fine_tuning"],
        ),
    ),
    ModelCard(
        model="gpt-4o-mini-2024-07-18",
        pricing=Pricing(text=TokenPrice(prompt=0.15, cached_discount=0.5, completion=0.60), audio=None, ft_price_hike=2.0),
        capabilities=ModelCapabilities(
            modalities=["text", "image"],
            endpoints=["chat_completions", "responses", "assistants", "batch", "fine_tuning"],
            features=["streaming", "function_calling", "structured_outputs", "fine_tuning"],
        ),
    ),
    # ----------------------
    # gpt-4o-mini-audio family
    # ----------------------
    ModelCard(
        model="gpt-4o-mini-audio-preview-2024-12-17",
        pricing=Pricing(
            text=TokenPrice(prompt=0.15, cached_discount=0.5, completion=0.60), audio=TokenPrice(prompt=10.00, cached_discount=0.2, completion=20.00), ft_price_hike=2.0
        ),
        capabilities=ModelCapabilities(modalities=["text", "audio"], endpoints=["chat_completions"], features=["streaming", "function_calling"]),
    ),
    ModelCard(
        model="gpt-4o-mini-realtime-preview-2024-12-17",
        pricing=Pricing(
            text=TokenPrice(prompt=0.60, cached_discount=0.5, completion=2.40), audio=TokenPrice(prompt=10.00, cached_discount=0.2, completion=20.00), ft_price_hike=2.0
        ),
        capabilities=ModelCapabilities(modalities=["text", "audio"], endpoints=["chat_completions"], features=["streaming", "function_calling"]),
    ),
]

# Anthropic Model Cards
anthropic_model_cards = [
    ModelCard(
        model="claude-3-5-sonnet-20241022",
        pricing=Pricing(text=TokenPrice(prompt=3.00, cached_discount=0.5, completion=15.00), audio=None),
        capabilities=ModelCapabilities(modalities=["text", "image"], endpoints=["chat_completions"], features=["streaming", "function_calling"]),
    ),
    ModelCard(
        model="claude-3-5-haiku-20241022",
        pricing=Pricing(text=TokenPrice(prompt=0.80, cached_discount=0.5, completion=4.00), audio=None),
        capabilities=ModelCapabilities(modalities=["text"], endpoints=["chat_completions"], features=["streaming", "function_calling"]),
    ),
    ModelCard(
        model="claude-3-opus-20240229",
        pricing=Pricing(text=TokenPrice(prompt=15.00, cached_discount=0.5, completion=75.00), audio=None),
        capabilities=ModelCapabilities(modalities=["text", "image"], endpoints=["chat_completions"], features=["streaming", "function_calling"]),
    ),
    ModelCard(
        model="claude-3-sonnet-20240229",
        pricing=Pricing(text=TokenPrice(prompt=3.00, cached_discount=0.5, completion=15.00), audio=None),
        capabilities=ModelCapabilities(modalities=["text", "image"], endpoints=["chat_completions"], features=["streaming", "function_calling"]),
    ),
    ModelCard(
        model="claude-3-haiku-20240307",
        pricing=Pricing(text=TokenPrice(prompt=0.25, cached_discount=0.5, completion=1.25), audio=None),
        capabilities=ModelCapabilities(modalities=["text", "image"], endpoints=["chat_completions"], features=["streaming", "function_calling"]),
    ),
]

# xAI model cards
xai_model_cards = [
    ModelCard(
        model="grok-2-vision-1212",
        pricing=Pricing(text=TokenPrice(prompt=2.00, cached_discount=0.5, completion=10.00), audio=None),
        capabilities=ModelCapabilities(modalities=["text", "image"], endpoints=["chat_completions"], features=["streaming", "function_calling"]),
    ),
    ModelCard(
        model="grok-2-1212",
        pricing=Pricing(text=TokenPrice(prompt=2.00, cached_discount=0.5, completion=10.00), audio=None),
        capabilities=ModelCapabilities(modalities=["text"], endpoints=["chat_completions"], features=["streaming", "function_calling"]),
    ),
    ModelCard(
        model="grok-3-beta",
        pricing=Pricing(text=TokenPrice(prompt=3.00, cached_discount=0.5, completion=15.00), audio=None),
        capabilities=ModelCapabilities(modalities=["text"], endpoints=["chat_completions"], features=["streaming", "function_calling"]),
    ),
    ModelCard(
        model="grok-3-mini-beta",
        pricing=Pricing(text=TokenPrice(prompt=0.30, cached_discount=0.5, completion=0.50), audio=None),
        capabilities=ModelCapabilities(modalities=["text"], endpoints=["chat_completions"], features=["streaming", "function_calling"]),
    ),
]

# Add Gemini model cards
gemini_model_cards = [
    # ----------------------
    # gemini-2.5-pro-exp-03-25 family
    # ----------------------
    ModelCard(
        model="gemini-2.5-pro-exp-03-25",
        pricing=Pricing(text=TokenPrice(prompt=1.25, cached_discount=0.25, completion=10.00), audio=None),
        capabilities=ModelCapabilities(modalities=["text", "image"], endpoints=["chat_completions"], features=["streaming", "function_calling", "structured_outputs"]),
        temperature_support=True,
    ),
    ModelCard(
        model="gemini-2.5-pro-preview-05-06",
        pricing=Pricing(text=TokenPrice(prompt=1.25, cached_discount=0.25, completion=10.00), audio=None),
        capabilities=ModelCapabilities(modalities=["text", "image"], endpoints=["chat_completions"], features=["streaming", "function_calling", "structured_outputs"]),
    ),
    ModelCard(
        model="gemini-2.5-pro-preview-03-25",
        pricing=Pricing(text=TokenPrice(prompt=1.25, cached_discount=0.25, completion=10.00), audio=None),
        capabilities=ModelCapabilities(modalities=["text", "image"], endpoints=["chat_completions"], features=["streaming", "function_calling", "structured_outputs"]),
        temperature_support=True,
    ),
    ModelCard(
        model="gemini-2.5-flash-preview-04-17",
        pricing=Pricing(text=TokenPrice(prompt=0.15, cached_discount=0.25, completion=0.60), audio=None),
        capabilities=ModelCapabilities(modalities=["text", "image"], endpoints=["chat_completions"], features=["streaming", "function_calling", "structured_outputs"]), 
        temperature_support=True,
    ),
    ModelCard(
        model="gemini-2.5-flash-preview-05-20",
        pricing=Pricing(text=TokenPrice(prompt=0.15, cached_discount=0.25, completion=0.60), audio=None),
        capabilities=ModelCapabilities(modalities=["text", "image"], endpoints=["chat_completions"], features=["streaming", "function_calling", "structured_outputs"]),
        temperature_support=True,
    ),
    # ----------------------
    # gemini-2.0-flash family
    # ----------------------
    ModelCard(
        model="gemini-2.0-flash",
        pricing=Pricing(text=TokenPrice(prompt=0.1, cached_discount=0.025 / 0.1, completion=0.40), audio=TokenPrice(prompt=0.7, cached_discount=0.175 / 0.7, completion=1000)),
        capabilities=ModelCapabilities(modalities=["text", "image"], endpoints=["chat_completions"], features=["streaming", "function_calling", "structured_outputs"]),
        temperature_support=True,
    ),
    ModelCard(
        model="gemini-2.0-flash-lite",
        pricing=Pricing(text=TokenPrice(prompt=0.075, cached_discount=1.0, completion=0.30), audio=TokenPrice(prompt=0.075, cached_discount=1.0, completion=1000)),
        capabilities=ModelCapabilities(
            modalities=["text", "image", "audio"],
            endpoints=["chat_completions"],
            features=["streaming", "structured_outputs"],
        ),
        temperature_support=True,
    ),
]


model_cards = openai_model_cards + gemini_model_cards + xai_model_cards
xAI_model_cards = [
    # ----------------------
    # grok3-family
    # ----------------------
    ModelCard(
        model="grok-3-beta",
        pricing=Pricing(text=TokenPrice(prompt=3, cached_discount=1.0, completion=15), audio=None),
        capabilities=ModelCapabilities(
            modalities=["text"],
            endpoints=["chat_completions"],
            features=["streaming", "structured_outputs"],
        ),
        temperature_support=True,
    ),
    ModelCard(
        model="grok-3-mini-beta",
        pricing=Pricing(text=TokenPrice(prompt=0.3, cached_discount=1.0, completion=0.5), audio=None),
        capabilities=ModelCapabilities(
            modalities=["text"],
            endpoints=["chat_completions"],
            features=["streaming", "structured_outputs"],
        ),
        temperature_support=True,
    ),
]


model_cards = openai_model_cards + gemini_model_cards + xAI_model_cards


def get_model_card(model: str) -> ModelCard:
    """
    Get the model card for a specific model.

    Args:
        model: The model name to look up

    Returns:
        The ModelCard for the specified model

    Raises:
        ValueError: If no model card is found for the specified model
    """
    # Extract base model name for fine-tuned models like "ft:gpt-4o:uiform:4389573"
    if model.startswith("ft:"):
        # Split by colon and take the second part (index 1) which contains the base model
        parts = model.split(":")
        if len(parts) > 1:
            model = parts[1]

    for card in model_cards:
        if card.model == model:
            return card

    raise ValueError(f"No model card found for model: {model}")



from io import IOBase
from pathlib import Path
from typing import Any, Literal

import PIL.Image
from pydantic import HttpUrl

from ..._resource import AsyncAPIResource, SyncAPIResource
from ..._utils.mime import convert_mime_data_to_pil_image, prepare_mime_document
from ...types.documents.create_messages import DocumentCreateMessageRequest, DocumentMessage, DocumentCreateInputRequest
from ...types.mime import MIMEData
from ...types.modalities import Modality
from ...types.standards import PreparedRequest
from .extractions import AsyncExtractions, Extractions
from ..._utils.json_schema import load_json_schema


class BaseDocumentsMixin:
    def _prepare_create_messages(
        self,
        document: Path | str | IOBase | MIMEData | PIL.Image.Image | HttpUrl,
        modality: Modality = "native",
        image_resolution_dpi: int | None = None,
        browser_canvas: Literal['A3', 'A4', 'A5'] | None = None,
        idempotency_key: str | None = None,
    ) -> PreparedRequest:
        mime_document = prepare_mime_document(document)
        data: dict[str, Any] = {
            "document": mime_document.model_dump(),
            "modality": modality,
        }
        if image_resolution_dpi:
            data["image_resolution_dpi"] = image_resolution_dpi
        if browser_canvas:
            data["browser_canvas"] = browser_canvas

        loading_request = DocumentCreateMessageRequest.model_validate(data)
        return PreparedRequest(method="POST", url="/v1/documents/create_messages", data=loading_request.model_dump(), idempotency_key=idempotency_key)

    def _prepare_create_inputs(
        self,
        document: Path | str | IOBase | MIMEData | PIL.Image.Image | HttpUrl,
        json_schema: dict[str, Any] | Path | str,
        modality: Modality = "native",
        image_resolution_dpi: int | None = None,
        browser_canvas: Literal['A3', 'A4', 'A5'] | None = None,
        idempotency_key: str | None = None,
    ) -> PreparedRequest:
        mime_document = prepare_mime_document(document)
        loaded_schema = load_json_schema(json_schema)

        data: dict[str, Any] = {
            "document": mime_document.model_dump(),
            "modality": modality,
            "json_schema": loaded_schema,
        }
        if image_resolution_dpi:
            data["image_resolution_dpi"] = image_resolution_dpi
        if browser_canvas:
            data["browser_canvas"] = browser_canvas

        loading_request = DocumentCreateInputRequest.model_validate(data)
        return PreparedRequest(method="POST", url="/v1/documents/create_inputs", data=loading_request.model_dump(), idempotency_key=idempotency_key)

    def _prepare_correct_image_orientation(self, document: Path | str | IOBase | MIMEData | PIL.Image.Image) -> PreparedRequest:
        mime_document = prepare_mime_document(document)

        if not mime_document.mime_type.startswith("image/"):
            raise ValueError("Image is not a valid image")

        return PreparedRequest(
            method="POST",
            url="/v1/documents/correct_image_orientation",
            data={"document": mime_document.model_dump()},
        )


class Documents(SyncAPIResource, BaseDocumentsMixin):
    """Documents API wrapper"""

    def __init__(self, client: Any) -> None:
        super().__init__(client=client)
        self.extractions = Extractions(client=client)
        # self.batch = Batch(client=client)

    def correct_image_orientation(self, document: Path | str | IOBase | MIMEData | PIL.Image.Image) -> PIL.Image.Image:
        """Corrects the orientation of an image using the UiForm API.

        This method takes an image in various formats and returns a PIL Image with corrected orientation.
        Useful for handling images from mobile devices or cameras that may have incorrect EXIF orientation.

        Args:
            image: The input image to correct. Can be:
                - A file path (Path or str)
                - A file-like object (IOBase)
                - A MIMEData object
                - A PIL Image object

        Returns:
            PIL.Image.Image: The orientation-corrected image as a PIL Image object

        Raises:
            ValueError: If the input is not a valid image
            UiformAPIError: If the API request fails
        """
        request = self._prepare_correct_image_orientation(document)
        response = self._client._prepared_request(request)
        mime_response = MIMEData.model_validate(response['document'])
        return convert_mime_data_to_pil_image(mime_response)

    def create_messages(
        self,
        document: Path | str | IOBase | MIMEData | PIL.Image.Image | HttpUrl,
        modality: Modality = "native",
        image_resolution_dpi: int | None = None,
        browser_canvas: Literal['A3', 'A4', 'A5'] | None = None,
        idempotency_key: str | None = None,
    ) -> DocumentMessage:
        """
        Create document messages from a file using the UiForm API.

        Args:
            document: The document to process. Can be a file path (Path or str) or a file-like object.
            modality: The processing modality to use. Defaults to "native".
            image_resolution_dpi: Optional image resolution DPI.
            browser_canvas: Optional browser canvas size.
            idempotency_key: Optional idempotency key for the request
        Returns:
            DocumentMessage: The processed document message containing extracted content.

        Raises:
            UiformAPIError: If the API request fails.
        """
        request = self._prepare_create_messages(document, modality, image_resolution_dpi, browser_canvas, idempotency_key)
        response = self._client._prepared_request(request)
        return DocumentMessage.model_validate(response)

    def create_inputs(
        self,
        document: Path | str | IOBase | MIMEData | PIL.Image.Image | HttpUrl,
        json_schema: dict[str, Any] | Path | str,
        modality: Modality = "native",
        image_resolution_dpi: int | None = None,
        browser_canvas: Literal['A3', 'A4', 'A5'] | None = None,
        idempotency_key: str | None = None,
    ) -> DocumentMessage:
        """
        Create document inputs (messages with schema) from a file using the UiForm API.

        Args:
            document: The document to process. Can be a file path (Path or str), file-like object, MIMEData, PIL Image, or URL.
            json_schema: The JSON schema to use for structuring the document content.
            modality: The processing modality to use. Defaults to "native".
            image_resolution_dpi: Optional image resolution DPI.
            browser_canvas: Optional browser canvas size.
            idempotency_key: Optional idempotency key for the request
        Returns:
            DocumentMessage: The processed document message containing extracted content with schema context.

        Raises:
            UiformAPIError: If the API request fails.
        """
        request = self._prepare_create_inputs(document, json_schema, modality, image_resolution_dpi, browser_canvas, idempotency_key)
        response = self._client._prepared_request(request)
        return DocumentMessage.model_validate(response)


class AsyncDocuments(AsyncAPIResource, BaseDocumentsMixin):
    """Documents API wrapper for asynchronous usage."""

    def __init__(self, client: Any) -> None:
        super().__init__(client=client)
        self.extractions = AsyncExtractions(client=client)

    async def create_messages(
        self,
        document: Path | str | IOBase | MIMEData | PIL.Image.Image,
        modality: Modality = "native",
        image_resolution_dpi: int | None = None,
        browser_canvas: Literal['A3', 'A4', 'A5'] | None = None,
        idempotency_key: str | None = None,
    ) -> DocumentMessage:
        """
        Create document messages from a file using the UiForm API asynchronously.

        Args:
            document: The document to process. Can be a file path (Path or str) or a file-like object.
            modality: The processing modality to use. Defaults to "native".
            idempotency_key: Idempotency key for request
        Returns:
            DocumentMessage: The processed document message containing extracted content.

        Raises:
            UiformAPIError: If the API request fails.
        """
        request = self._prepare_create_messages(document, modality, image_resolution_dpi, browser_canvas, idempotency_key)
        assert request.data is not None
        print(request.data.keys())
        print(request.data)
        response = await self._client._prepared_request(request)
        return DocumentMessage.model_validate(response)

    async def create_inputs(
        self,
        document: Path | str | IOBase | MIMEData | PIL.Image.Image | HttpUrl,
        json_schema: dict[str, Any] | Path | str,
        modality: Modality = "native",
        image_resolution_dpi: int | None = None,
        browser_canvas: Literal['A3', 'A4', 'A5'] | None = None,
        idempotency_key: str | None = None,
    ) -> DocumentMessage:
        """
        Create document inputs (messages with schema) from a file using the UiForm API asynchronously.

        Args:
            document: The document to process. Can be a file path (Path or str), file-like object, MIMEData, PIL Image, or URL.
            json_schema: The JSON schema to use for structuring the document content.
            modality: The processing modality to use. Defaults to "native".
            image_resolution_dpi: Optional image resolution DPI.
            browser_canvas: Optional browser canvas size.
            idempotency_key: Idempotency key for request
        Returns:
            DocumentMessage: The processed document message containing extracted content with schema context.

        Raises:
            UiformAPIError: If the API request fails.
        """
        request = self._prepare_create_inputs(document, json_schema, modality, image_resolution_dpi, browser_canvas, idempotency_key)
        response = await self._client._prepared_request(request)
        return DocumentMessage.model_validate(response)

    async def correct_image_orientation(self, document: Path | str | IOBase | MIMEData | PIL.Image.Image) -> PIL.Image.Image:
        """Corrects the orientation of an image using the UiForm API asynchronously.

        This method takes an image in various formats and returns a PIL Image with corrected orientation.
        Useful for handling images from mobile devices or cameras that may have incorrect EXIF orientation.

        Args:
            image: The input image to correct. Can be:
                - A file path (Path or str)
                - A file-like object (IOBase)
                - A MIMEData object
                - A PIL Image object

        Returns:
            PIL.Image.Image: The orientation-corrected image as a PIL Image object

        Raises:
            ValueError: If the input is not a valid image
            UiformAPIError: If the API request fails
        """
        request = self._prepare_correct_image_orientation(document)
        response = await self._client._prepared_request(request)
        mime_response = MIMEData.model_validate(response['document'])
        return convert_mime_data_to_pil_image(mime_response)



from pydantic import BaseModel

from ..._utils.benchmarking import EvalMetrics, SingleFileEval, compute_dict_difference
from .batch_annotation import AnnotationInputData, InferenceSettings

# This job will generate two datasets from the original dataset, one with the first annotation and one with the second annotation
# It will then evaluate the two datasets using the evaluation metrics and return an EvalMetrics object


class EvaluationInputData(BaseModel):
    original_dataset_id: str
    schema_id: str
    schema_data_id: str
    inference_settings_1: InferenceSettings
    inference_settings_2: InferenceSettings


# def evaluate_datasets(
#     original_dataset_id: str,
#     inference_settings_1: InferenceSettings,
#     inference_settings_2: InferenceSettings,
#     identity: Identity,
#     job_execution_id: str,
#     settings: Settings,
#     dashboard_db: AsyncIOMotorDatabase,
# ) -> EvalMetrics:
#     # Generate two datasets from the original dataset

#     # Create the actual dataset objects.

#     # Solution:
#     # 1. Create the two datasets objects
#     # 2. Duplicate all the dataset membership objects for the two datasets (with the right dataset_id)

#     # 3. Annotate the two datasets with the two annotation props
#     annotation_job_1 = AnnotationJob(
#         input_data=AnnotationInputData(
#             dataset_id=original_dataset_id,
#             files_ids=None,
#             upsert=True,
#             inference_settings=inference_settings_1
#         )
#     )

#     annotation_job_2 = AnnotationJob(
#         input_data=AnnotationInputData(
#             dataset_id=original_dataset_id,
#             files_ids=None,
#             upsert=True,
#             inference_settings=inference_settings_2
#         )
#     )
#     batch_annotate_job_with_checkpoints(
#         identity=identity,
#         job_execution_id=job_execution_id,
#         annotation_job=annotation_job_1,
#         settings=settings,
#         dashboard_db=dashboard_db,
#     )

#     batch_annotate_job_with_checkpoints(
#         identity=identity,
#         job_execution_id=job_execution_id,
#         annotation_job=annotation_job_2,
#         settings=settings,
#         dashboard_db=dashboard_db,
#     )

#     def compute_all_single_file_evals(
#         dataset_1: Dataset,
#         dataset_2: Dataset,
#     ) -> list[SingleFileEval]:

#         single_file_evals: list[SingleFileEval] = []
#         for file_id in dataset_1.file_ids:
#             single_file_evals.append(
#                 SingleFileEval(
#                     file_id=file_id,
#                     dict_1=dataset_1,
#                     dict_2=dataset_2.get_file(file_id),
#                 )
#             )

#         for file_id in dataset_2.file_ids:
#             single_file_evals.append(
#                 SingleFileEval(
#                     file_id=file_id,
#                     dict_1=dataset_2.get_file(file_id),
#                     dict_2=dataset_1,
#                 )
#             )

#         for file_id in dataset_1.file_ids:
#             single_file_evals.append(SingleFileEval(
#                 file_id=file_id,
#                 dict_1=dataset_1.get_file(file_id),
#                 dict_2=dataset_2.get_file(file_id),
#                 schema_id=schema_id,
#                 schema_data_id=schema_data_id,
#                 dataset_membership_id_1=dataset_1.get_file(file_id).id,
#                 dataset_membership_id_2=dataset_2.get_file(file_id).id,
#                 hamming_similarity=compute_dict_difference(
#                     dict_1=dataset_1.get_file(file_id),
#                     dict_2=dataset_2.get_file(file_id),
#                     metric="hamming_similarity"
#                 ),
#                 jaccard_similarity=compute_dict_difference(
#                     dict_1=dataset_1.get_file(file_id),
#                     dict_2=dataset_2.get_file(file_id),
#                     metric="jaccard_similarity"
#                 ),
#                 levenshtein_similarity=compute_dict_difference(
#                     dict_1=dataset_1.get_file(file_id),
#                     dict_2=dataset_2.get_file(file_id),
#                     metric="levenshtein_similarity"
#                 )
#                 )


#         )
#     # Then go through all the entries in the datasets and compute the evaluation metrics
#     compute_all_single_file_evals(
#         dataset_1=dataset_1,
#         dataset_2=dataset_2,
#     )
#     # Return the EvalMetrics object

#     compute_eval_metrics


#     raise NotImplementedError("Not implemented")

#     return eval_metrics



# from typing import Literal, Any
# from pydantic import BaseModel, computed_field
# from ..mime import MIMEData
# from ..._utils.benchmarking import ExtractionAnalysis

# MAX_CONCURRENCY = 15


# class PromptOptimizationObject(BaseModel):
#     mime_document: MIMEData
#     target: dict
#     extracted: dict | None = None

#     @computed_field     # type: ignore
#     @property
#     def analysis(self) -> ExtractionAnalysis | None:
#         if self.extracted is None:
#             return None
#         return ExtractionAnalysis(ground_truth=self.target, prediction=self.extracted)


# Metrics = Literal["levenshtein_similarity_per_field", "accuracy_per_field"]


# # Insert default values for the parameters
# class PromptOptimizationProps(BaseModel):
#     start_hierarchy_level: int = 0
#     threshold: float = 0.9
#     model: str = "gpt-4o-mini"
#     iterations_per_level: int = 1
#     metric: Metrics = "accuracy_per_field"

# class PromptOptimizationJobInputData(BaseModel):
#     json_schema: dict[str, Any]
#     optimization_objects: list[PromptOptimizationObject]
#     schema_optimization_props: PromptOptimizationProps


# class PromptOptimizationJob(BaseModel):
#     job_type: Literal["prompt-optimization"] = "prompt-optimization"
#     input_data: PromptOptimizationJobInputData



from typing import Any, Literal

from openai import AsyncOpenAI, OpenAI

from .._resource import AsyncAPIResource, SyncAPIResource


class Files(SyncAPIResource):
    """Files API wrapper"""

    # Tell Pydantic that _client is not a "field" for serialization/validation,
    # but rather a private attribute to store arbitrary data.

    def create(self, file: Any, purpose: Literal['assistants', 'batch', 'fine-tune', 'vision']) -> Any:
        openai_client = OpenAI(api_key=self._client.headers["OpenAI-Api-Key"])
        return openai_client.files.create(file=file, purpose=purpose)


class AsyncFiles(AsyncAPIResource):
    """Files Asyncronous API wrapper"""

    async def create(self, file: Any, purpose: Literal['assistants', 'batch', 'fine-tune', 'vision']) -> Any:
        async with AsyncOpenAI(api_key=self._client.headers["OpenAI-Api-Key"]) as openai_client:
            return await openai_client.files.create(file=file, purpose=purpose)



from datetime import datetime
from enum import Enum
from typing import Optional

from pydantic import BaseModel

from ..ai_models import AIProvider


class ExternalAPIKeyRequest(BaseModel):
    """Request model for creating/updating API keys"""

    provider: AIProvider
    api_key: str


class ExternalAPIKey(BaseModel):
    """Response model for API key information"""

    provider: AIProvider
    is_configured: bool
    last_updated: Optional[datetime]



import os
from typing import List

from ..._resource import AsyncAPIResource, SyncAPIResource
from ...types.ai_models import AIProvider
from ...types.secrets.external_api_keys import ExternalAPIKey, ExternalAPIKeyRequest
from ...types.standards import PreparedRequest


class ExternalAPIKeysMixin:
    def prepare_create(self, provider: AIProvider, api_key: str) -> PreparedRequest:
        data = {"provider": provider, "api_key": api_key}
        request = ExternalAPIKeyRequest.model_validate(data)
        return PreparedRequest(method="POST", url="/v1/secrets/external_api_keys", data=request.model_dump(mode="json"))

    def prepare_get(self, provider: AIProvider) -> PreparedRequest:
        return PreparedRequest(method="GET", url=f"/v1/secrets/external_api_keys/{provider}")

    def prepare_list(self) -> PreparedRequest:
        return PreparedRequest(method="GET", url="/v1/secrets/external_api_keys")

    def prepare_delete(self, provider: AIProvider) -> PreparedRequest:
        return PreparedRequest(method="DELETE", url=f"/v1/secrets/external_api_keys/{provider}")


class ExternalAPIKeys(SyncAPIResource, ExternalAPIKeysMixin):
    """External API Keys management wrapper"""

    def create(self, provider: AIProvider, api_key: str) -> dict:
        """Add or update an external API key.

        Args:
            provider: The API provider (openai, gemini, anthropic, xai)
            api_key: The API key to store

        Returns:
            dict: Response indicating success
        """

        request = self.prepare_create(provider, api_key)
        response = self._client._prepared_request(request)
        return response

    def get(
        self,
        provider: AIProvider,
    ) -> ExternalAPIKey:
        """Get an external API key configuration.

        Args:
            provider: The API provider to get the key for

        Returns:
            ExternalAPIKey: The API key configuration
        """
        request = self.prepare_get(provider)
        response = self._client._prepared_request(request)
        return response

        return ExternalAPIKey.model_validate(response)

    def list(self) -> List[ExternalAPIKey]:
        """List all configured external API keys.

        Returns:
            List[ExternalAPIKey]: List of API key configurations
        """
        request = self.prepare_list()
        response = self._client._prepared_request(request)

        return [ExternalAPIKey.model_validate(key) for key in response]

    def delete(self, provider: AIProvider) -> dict:
        """Delete an external API key configuration.

        Args:
            provider: The API provider to delete the key for

        Returns:
            dict: Response indicating success
        """
        request = self.prepare_delete(provider)
        response = self._client._prepared_request(request)

        return response


class AsyncExternalAPIKeys(AsyncAPIResource, ExternalAPIKeysMixin):
    """External API Keys management wrapper"""

    async def create(self, provider: AIProvider, api_key: str) -> dict:
        request = self.prepare_create(provider, api_key)
        response = await self._client._prepared_request(request)
        return response

    async def get(self, provider: AIProvider) -> ExternalAPIKey:
        request = self.prepare_get(provider)
        response = await self._client._prepared_request(request)
        return response

    async def list(self) -> List[ExternalAPIKey]:
        request = self.prepare_list()
        response = await self._client._prepared_request(request)
        return [ExternalAPIKey.model_validate(key) for key in response]

    async def delete(self, provider: AIProvider) -> dict:
        request = self.prepare_delete(provider)
        response = await self._client._prepared_request(request)
        return response



import re
import unicodedata
from collections import defaultdict
from typing import Any, Literal, Optional

import numpy as np
import pandas as pd
import termplotlib as tpl  # type: ignore
from Levenshtein import distance as levenshtein_distance
from pydantic import BaseModel, computed_field

# The goal is to leverage this piece of code to open a jsonl file and get an analysis of the performance of the model using a one-liner.


############# BENCHMARKING MODELS #############


class DictionaryComparisonMetrics(BaseModel):
    # Pure dict comparison
    unchanged_fields: int
    total_fields: int
    is_equal: dict[str, bool]
    false_positives: list[dict[str, Any]]
    false_negatives: list[dict[str, Any]]
    mismatched_values: list[dict[str, Any]]
    keys_only_on_1: list[str]
    keys_only_on_2: list[str]

    # Some metrics
    valid_comparisons: int
    total_accuracy: float
    false_positive_rate: float
    false_negative_rate: float
    mismatched_value_rate: float

    similarity_levenshtein: dict[str, float]
    similarity_jaccard: dict[str, float]

    avg_similarity_levenshtein: float
    avg_similarity_jaccard: float
    total_similarity_levenshtein: float
    total_similarity_jaccard: float


def flatten_dict(obj: Any, prefix: str = '') -> dict[str, Any]:
    items = []  # type: ignore
    if isinstance(obj, dict):
        for k, v in obj.items():
            new_key = f"{prefix}.{k}" if prefix else k
            items.extend(flatten_dict(v, new_key).items())
    elif isinstance(obj, list):
        for i, v in enumerate(obj):
            new_key = f"{prefix}.{i}"
            items.extend(flatten_dict(v, new_key).items())
    else:
        items.append((prefix, obj))
    return dict(items)


def normalize_value(val: Any) -> str:
    """Convert value to uppercase and remove all spacing for comparison."""
    if val is None:
        return ""
    prep = re.sub(r'\s+', '', str(val).upper())
    # Remove all accents (é -> e, etc.)
    return unicodedata.normalize('NFKD', prep).encode('ASCII', 'ignore').decode()


def key_normalization(key: str) -> str:
    """This method is useful to compare keys under list indexes (that refers to the same kind of error but on different list index position)"""
    # We will replace all .{i} with .* where i is the index of the list (using regex for this)
    key_parts = key.split(".")
    new_key_parts = []
    for key_part in key_parts:
        if key_part.isdigit():
            new_key_parts.append("*")
        else:
            new_key_parts.append(key_part)
    return ".".join(new_key_parts)


def should_ignore_key(
    key: str, exclude_field_patterns: list[str] | None, include_field_patterns: list[str] | None = None, information_presence_per_field: dict[str, bool] | None = None
) -> bool:
    if information_presence_per_field and information_presence_per_field.get(key) is False:
        # If we have the information_presence_per_field dict and the key is marked as false, then we should ignore it
        should_ignore = True
    else:
        # If exclude_field_patterns is None, we should not ignore any key
        normalized_key = key_normalization(key)
        should_ignore = any(normalized_key.startswith(key_normalization(pattern)) for pattern in exclude_field_patterns or [])

        if include_field_patterns and not should_ignore:
            # If include_field_patterns is not None, we should ignore the key if it does not start with any of the include_field_patterns and is not in the exclude_field_patterns
            should_ignore = not any(normalized_key.startswith(key_normalization(pattern)) for pattern in include_field_patterns)

    return should_ignore


def levenshtein_similarity(val1: Any, val2: Any) -> float:
    """
    Calculate similarity between two values using Levenshtein distance.
    Returns a similarity score between 0.0 and 1.0.
    """
    # Handle None/empty and general cases
    if (val1 or "") == (val2 or ""):
        return 1.0

    # Check if both values are numeric, compare with 5% tolerance
    if isinstance(val1, (int, float)) and isinstance(val2, (int, float)):
        return 1.0 if abs(val1 - val2) <= 0.05 * max(abs(val1), abs(val2)) else 0.0

    # Convert to normalized strings
    str1 = normalize_value(val1)
    str2 = normalize_value(val2)

    if str1 == str2:
        return 1.0

    # Calculate Levenshtein distance
    if str1 and str2:  # Only if both strings are non-empty
        max_len = max(len(str1), len(str2))
        if max_len == 0:
            return 1.0

        dist = levenshtein_distance(str1, str2)
        return 1 - (dist / max_len)

    return 0.0


def jaccard_similarity(val1: Any, val2: Any) -> float:
    """
    Calculate Jaccard similarity between two values.
    Returns a similarity score between 0.0 and 1.0.
    """
    # Handle None/empty and general cases
    if (val1 or "") == (val2 or ""):
        return 1.0

    # Check if both values are numeric, compare with 5% tolerance
    if isinstance(val1, (int, float)) and isinstance(val2, (int, float)):
        return 1.0 if abs(val1 - val2) <= 0.05 * max(abs(val1), abs(val2)) else 0.0

    # Convert to normalized strings and split into words
    str1 = set(normalize_value(val1).split())
    str2 = set(normalize_value(val2).split())

    if not str1 and not str2:
        return 1.0

    # Calculate Jaccard similarity
    intersection = len(str1.intersection(str2))
    union = len(str1.union(str2))

    return intersection / union if union > 0 else 0.0


def compare_dicts(
    ground_truth: dict[str, Any],
    prediction: dict[str, Any],
    include_fields: list[str] | None = None,
    exclude_fields: list[str] | None = None,
    information_presence_per_field: dict[str, bool] | None = None,
    levenshtein_threshold: float = 0.0,  # 0.0 means exact match
) -> DictionaryComparisonMetrics:
    flat_ground_truth = flatten_dict(ground_truth)
    flat_prediction = flatten_dict(prediction)

    flat_ground_truth = {k: v for k, v in flat_ground_truth.items() if not should_ignore_key(k, exclude_fields, include_fields, information_presence_per_field)}
    flat_prediction = {k: v for k, v in flat_prediction.items() if not should_ignore_key(k, exclude_fields, include_fields, information_presence_per_field)}

    keys_ground_truth = set(flat_ground_truth.keys())
    keys_prediction = set(flat_prediction.keys())
    common_keys = keys_ground_truth & keys_prediction

    keys_only_on_1 = sorted(list(keys_ground_truth - keys_prediction))
    keys_only_on_2 = sorted(list(keys_prediction - keys_ground_truth))

    total_fields = len(common_keys)
    unchanged_fields = 0
    is_equal_per_field = {}

    false_positives = []
    false_negatives = []
    mismatched_values = []

    total_similarity_levenshtein = 0.0
    total_similarity_jaccard = 0.0
    similarity_levenshtein_per_field = {}
    similarity_jaccard_per_field = {}

    valid_comparisons = 0

    for key in common_keys:
        llm_value = flat_ground_truth[key]
        extraction_value = flat_prediction[key]

        coerced_llm_value = llm_value or ""
        coerced_extraction_value = extraction_value or ""

        similarity_lev = levenshtein_similarity(llm_value, extraction_value)
        similarity_jac = jaccard_similarity(llm_value, extraction_value)
        # print("Jaccard similarity", similarity_jac)

        # Use Levenshtein for equality comparison (you can adjust this if needed)
        is_equal = similarity_lev >= (1 - levenshtein_threshold)

        similarity_levenshtein_per_field[key] = similarity_lev
        similarity_jaccard_per_field[key] = similarity_jac
        is_equal_per_field[key] = is_equal

        # Only count non-empty comparisons for average similarity
        if coerced_llm_value != "" and coerced_extraction_value != "":
            total_similarity_levenshtein += similarity_lev
            total_similarity_jaccard += similarity_jac
            valid_comparisons += 1

        if is_equal:
            unchanged_fields += 1
        else:
            if coerced_llm_value != "" and coerced_extraction_value == "":
                false_positives.append({"key": key, "expected": extraction_value, "got": llm_value, "similarity": similarity_lev})
            elif coerced_llm_value == "" and coerced_extraction_value != "":
                false_negatives.append({"key": key, "expected": extraction_value, "got": llm_value, "similarity": similarity_lev})
            elif coerced_llm_value != "" and coerced_extraction_value != "":
                # Both are non-empty but not equal
                mismatched_values.append({"key": key, "expected": extraction_value, "got": llm_value, "similarity": similarity_lev})
    # Some metrics
    avg_similarity_levenshtein = total_similarity_levenshtein / valid_comparisons if valid_comparisons > 0 else 1.0
    avg_similarity_jaccard = total_similarity_jaccard / valid_comparisons if valid_comparisons > 0 else 1.0
    total_accuracy = unchanged_fields / total_fields if total_fields > 0 else 1.0
    false_positive_rate = len(false_positives) / total_fields if total_fields > 0 else 0.0
    false_negative_rate = len(false_negatives) / total_fields if total_fields > 0 else 0.0
    mismatched_value_rate = len(mismatched_values) / total_fields if total_fields > 0 else 0.0

    return DictionaryComparisonMetrics(
        unchanged_fields=unchanged_fields,
        total_fields=total_fields,
        is_equal=is_equal_per_field,
        false_positives=false_positives,
        false_negatives=false_negatives,
        mismatched_values=mismatched_values,
        keys_only_on_1=keys_only_on_1,
        keys_only_on_2=keys_only_on_2,
        valid_comparisons=valid_comparisons,
        total_accuracy=total_accuracy,
        false_positive_rate=false_positive_rate,
        false_negative_rate=false_negative_rate,
        mismatched_value_rate=mismatched_value_rate,
        similarity_levenshtein=similarity_levenshtein_per_field,
        similarity_jaccard=similarity_jaccard_per_field,
        avg_similarity_levenshtein=avg_similarity_levenshtein,
        avg_similarity_jaccard=avg_similarity_jaccard,
        total_similarity_levenshtein=total_similarity_levenshtein,
        total_similarity_jaccard=total_similarity_jaccard,
    )


class ExtractionAnalysis(BaseModel):
    ground_truth: dict[str, Any]
    prediction: dict[str, Any]
    time_spent: Optional[float] = None
    include_fields: list[str] | None = None
    exclude_fields: list[str] | None = None
    information_presence_per_field: dict[str, bool] | None = None
    levenshtein_threshold: float = 0.0

    @computed_field  # type: ignore
    @property
    def comparison(self) -> DictionaryComparisonMetrics:
        return compare_dicts(
            self.ground_truth,
            self.prediction,
            include_fields=self.include_fields,
            exclude_fields=self.exclude_fields,
            information_presence_per_field=self.information_presence_per_field,
            levenshtein_threshold=self.levenshtein_threshold,
        )


class BenchmarkMetrics(BaseModel):
    ai_model: str
    accuracy: float
    levenshtein_similarity: float
    jaccard_similarity: float
    false_positive_rate: float
    false_negative_rate: float
    mismatched_value_rate: float


from rich.console import Console
from rich.table import Table


def display_benchmark_metrics(benchmark_metrics: list[BenchmarkMetrics]) -> None:
    """
    Display benchmark metrics for multiple models in a formatted table.

    Args:
        benchmark_metrics: List of BenchmarkMetrics objects containing model performance data
    """
    console = Console(style="on grey23")
    table = Table(title="Model Benchmark Comparison", show_lines=True)

    # Add columns
    table.add_column("Model", justify="left", style="#BDE8F6", no_wrap=True)
    table.add_column("Accuracy", justify="right", style="#C2BDF6")
    table.add_column("Levenshtein", justify="right", style="#F6BDBD")
    table.add_column("Jaccard", justify="right", style="#F6E4BD")
    table.add_column("False Positive Rate", justify="right", style="#BDF6C0")
    table.add_column("False Negative Rate", justify="right", style="#F6BDE4")
    table.add_column("Mismatched Value Rate", justify="right", style="#E4F6BD")

    # Find best values for each metric
    best_values = {
        'accuracy': max(m.accuracy for m in benchmark_metrics),
        'levenshtein': max(m.levenshtein_similarity for m in benchmark_metrics),
        'jaccard': max(m.jaccard_similarity for m in benchmark_metrics),
        'fp_rate': min(m.false_positive_rate for m in benchmark_metrics),
        'fn_rate': min(m.false_negative_rate for m in benchmark_metrics),
        'mismatch_rate': min(m.mismatched_value_rate for m in benchmark_metrics),
    }

    # Add rows for each model's metrics
    for metrics in benchmark_metrics:
        table.add_row(
            metrics.ai_models,
            f"[bold]{metrics.accuracy:.3f}[/bold]" if metrics.accuracy == best_values['accuracy'] else f"[dim]{metrics.accuracy:.3f}[/dim]",
            f"[bold]{metrics.levenshtein_similarity:.3f}[/bold]"
            if metrics.levenshtein_similarity == best_values['levenshtein']
            else f"[dim]{metrics.levenshtein_similarity:.3f}[/dim]",
            f"[bold]{metrics.jaccard_similarity:.3f}[/bold]" if metrics.jaccard_similarity == best_values['jaccard'] else f"[dim]{metrics.jaccard_similarity:.3f}[/dim]",
            f"[bold]{metrics.false_positive_rate:.3f}[/bold]" if metrics.false_positive_rate == best_values['fp_rate'] else f"[dim]{metrics.false_positive_rate:.3f}[/dim]",
            f"[bold]{metrics.false_negative_rate:.3f}[/bold]" if metrics.false_negative_rate == best_values['fn_rate'] else f"[dim]{metrics.false_negative_rate:.3f}[/dim]",
            f"[bold]{metrics.mismatched_value_rate:.3f}[/bold]"
            if metrics.mismatched_value_rate == best_values['mismatch_rate']
            else f"[dim]{metrics.mismatched_value_rate:.3f}[/dim]",
        )

    # Print the table
    console.print(table)


class ComparisonMetrics(BaseModel):
    # Total Values (count or sum) per Field
    false_positive_counts: dict[str, int] = defaultdict(int)
    false_positive_rate_per_field: dict[str, float] = defaultdict(float)

    false_negative_counts: dict[str, int] = defaultdict(int)
    false_negative_rate_per_field: dict[str, float] = defaultdict(float)

    mismatched_value_counts: dict[str, int] = defaultdict(int)
    mismatched_value_rate_per_field: dict[str, float] = defaultdict(float)

    common_presence_counts: dict[str, int] = defaultdict(int)
    accuracy_per_field: dict[str, float] = defaultdict(float)

    jaccard_similarity_per_field: dict[str, float] = defaultdict(float)
    total_jaccard_similarity_per_field: dict[str, float] = defaultdict(float)

    levenshtein_similarity_per_field: dict[str, float] = defaultdict(float)
    total_levenshtein_similarity_per_field: dict[str, float] = defaultdict(float)

    @computed_field  # type: ignore
    @property
    def accuracy(self) -> float:
        return sum(self.accuracy_per_field.values()) / len(self.accuracy_per_field)

    @computed_field  # type: ignore
    @property
    def levenshtein_similarity(self) -> float:
        return sum(self.levenshtein_similarity_per_field.values()) / len(self.levenshtein_similarity_per_field)

    @computed_field  # type: ignore
    @property
    def jaccard_similarity(self) -> float:
        return sum(self.jaccard_similarity_per_field.values()) / len(self.jaccard_similarity_per_field)

    @computed_field  # type: ignore
    @property
    def false_positive_rate(self) -> float:
        return sum(self.false_positive_rate_per_field.values()) / len(self.false_positive_rate_per_field)

    @computed_field  # type: ignore
    @property
    def false_negative_rate(self) -> float:
        return sum(self.false_negative_rate_per_field.values()) / len(self.false_negative_rate_per_field)

    @computed_field  # type: ignore
    @property
    def mismatched_value_rate(self) -> float:
        return sum(self.mismatched_value_rate_per_field.values()) / len(self.mismatched_value_rate_per_field)


def normalized_comparison_metrics(list_analyses: list[ExtractionAnalysis], min_freq: float = 0.2) -> ComparisonMetrics:
    false_positive_counts: dict[str, int] = defaultdict(int)
    false_negative_counts: dict[str, int] = defaultdict(int)
    mismatched_value_counts: dict[str, int] = defaultdict(int)
    common_presence_counts: dict[str, int] = defaultdict(int)
    is_equal_per_field: dict[str, int] = defaultdict(int)

    total_levenshtein_similarity_per_field: dict[str, float] = defaultdict(float)
    total_jaccard_similarity_per_field: dict[str, float] = defaultdict(float)
    false_positive_rate_per_field: dict[str, float] = defaultdict(float)
    false_negative_rate_per_field: dict[str, float] = defaultdict(float)
    mismatched_value_rate_per_field: dict[str, float] = defaultdict(float)

    for analysis in list_analyses:
        # Count false positives
        for error in analysis.comparison.false_positives:
            key = error["key"]
            false_positive_counts[key_normalization(key)] += 1

        # Count false negatives
        for error in analysis.comparison.false_negatives:
            key = error["key"]
            false_negative_counts[key_normalization(key)] += 1

        # Count Wrong Predictions
        for error in analysis.comparison.mismatched_values:
            key = error["key"]
            mismatched_value_counts[key_normalization(key)] += 1

        # Count total errors per field (Levenshtein)
        for key, similarity in analysis.comparison.similarity_levenshtein.items():
            common_presence_counts[key_normalization(key)] += 1
            total_levenshtein_similarity_per_field[key_normalization(key)] += similarity

        for key, is_equal in analysis.comparison.is_equal.items():
            is_equal_per_field[key_normalization(key)] += int(is_equal)

        # Count Jaccard Similarity
        for key, similarity in analysis.comparison.similarity_jaccard.items():
            total_jaccard_similarity_per_field[key_normalization(key)] += similarity

    accuracy_per_field = {
        key: is_equal_per_field[key] / common_presence_counts[key] for key in common_presence_counts if common_presence_counts[key] > int(min_freq * len(list_analyses))
    }
    levenshtein_similarity_per_field = {
        key: total_levenshtein_similarity_per_field[key] / common_presence_counts[key]
        for key in common_presence_counts
        if common_presence_counts[key] > int(min_freq * len(list_analyses))
    }
    jaccard_similarity_per_field = {
        key: total_jaccard_similarity_per_field[key] / common_presence_counts[key]
        for key in common_presence_counts
        if common_presence_counts[key] > int(min_freq * len(list_analyses))
    }
    false_positive_rate_per_field = {
        key: false_positive_counts[key] / common_presence_counts[key] for key in common_presence_counts if common_presence_counts[key] > int(min_freq * len(list_analyses))
    }
    false_negative_rate_per_field = {
        key: false_negative_counts[key] / common_presence_counts[key] for key in common_presence_counts if common_presence_counts[key] > int(min_freq * len(list_analyses))
    }
    mismatched_value_rate_per_field = {
        key: mismatched_value_counts[key] / common_presence_counts[key] for key in common_presence_counts if common_presence_counts[key] > int(min_freq * len(list_analyses))
    }

    return ComparisonMetrics(
        false_positive_counts=false_positive_counts,
        false_negative_counts=false_negative_counts,
        mismatched_value_counts=mismatched_value_counts,
        common_presence_counts=common_presence_counts,
        total_levenshtein_similarity_per_field=total_levenshtein_similarity_per_field,
        total_jaccard_similarity_per_field=total_jaccard_similarity_per_field,
        accuracy_per_field=accuracy_per_field,
        levenshtein_similarity_per_field=levenshtein_similarity_per_field,
        jaccard_similarity_per_field=jaccard_similarity_per_field,
        false_positive_rate_per_field=false_positive_rate_per_field,
        false_negative_rate_per_field=false_negative_rate_per_field,
        mismatched_value_rate_per_field=mismatched_value_rate_per_field,
    )


def plot_metric(
    analysis: ComparisonMetrics,
    value_type: Literal["accuracy", "levenshtein_similarity", "jaccard_similarity", "false_positive_rate", "false_negative_rate", "mismatched_value_rate"] = "accuracy",
    top_n: int = 20,
    ascending: bool = False,
) -> None:
    """Plot a metric from analysis results using a horizontal bar chart.

    Args:
        analysis: ComparisonMetrics object containing the analysis results
    """
    # Create dataframe from accuracy data
    df = pd.DataFrame(list(analysis.__getattribute__(value_type + "_per_field").items()), columns=["field", value_type]).sort_values(by=value_type, ascending=ascending)

    # Filter top n fields with the lowest accuracy
    top_n_df = df.head(top_n)

    # Create the plot
    fig = tpl.figure()
    fig.barh(np.array(top_n_df[value_type]).round(4), np.array(top_n_df["field"]), force_ascii=False)

    fig.show()


def plot_comparison_metrics(analysis: ComparisonMetrics, top_n: int = 20) -> None:
    metric_ascendency_dict: dict[
        Literal["accuracy", "levenshtein_similarity", "jaccard_similarity", "false_positive_rate", "false_negative_rate", "mismatched_value_rate"], bool
    ] = {"accuracy": True, "levenshtein_similarity": True, "jaccard_similarity": True, "false_positive_rate": False, "false_negative_rate": False, "mismatched_value_rate": False}

    print(f"#########################################")
    print(f"############ AVERAGE METRICS ############")
    print(f"#########################################")
    print(f"Accuracy: {analysis.accuracy:.2f}")
    print(f"Levenshtein Similarity: {analysis.levenshtein_similarity:.2f}")
    print(f"Jaccard Similarity (IOU): {analysis.jaccard_similarity:.2f}")
    print(f"False Positive Rate: {analysis.false_positive_rate:.2f}")
    print(f"False Negative Rate: {analysis.false_negative_rate:.2f}")
    print(f"Mismatched Value Rate: {analysis.mismatched_value_rate:.2f}")

    for metric, ascending in metric_ascendency_dict.items():
        print(f"\n\n############ {metric.upper()} ############")
        plot_metric(analysis, metric, top_n, ascending)


def get_aggregation_metrics(metric: dict[str, float | int], _hierarchy_level: int) -> dict[str, float]:
    if _hierarchy_level == 0:
        # For level 0, aggregate all values under empty string key
        return {"": sum(metric.values()) / len(metric)}

    aggregated_metrics: dict[str, list[float | int]] = {}
    for key, value in metric.items():
        # Split key and handle array notation by replacing array indices with '*'
        key_parts: list[str] = []
        for part in key.split('.'):
            if part.isdigit():
                key_parts.append('*')
            else:
                key_parts.append(part)
        actual_depth = len([part for part in key_parts if part != '*'])
        if actual_depth < _hierarchy_level:
            continue

        used_depth = 0
        aggregation_prefix_parts: list[str] = []
        for part in key_parts:
            if part == "*":
                aggregation_prefix_parts.append("*")
            else:
                aggregation_prefix_parts.append(part)
                used_depth += 1
            if used_depth == _hierarchy_level:
                break
        aggregation_prefix = '.'.join(aggregation_prefix_parts)

        # Aggregate metrics
        if aggregation_prefix not in aggregated_metrics:
            aggregated_metrics[aggregation_prefix] = []
        aggregated_metrics[aggregation_prefix].append(value)

    # Calculate averages
    return {key: sum(values) / len(values) if len(values) > 0 else 0 for key, values in aggregated_metrics.items()}


def get_max_depth(metric: dict[str, float | int]) -> int:
    max_depth_upper_bound = max(len(key.split('.')) for key in metric.keys())
    for max_depth in range(max_depth_upper_bound, 0, -1):
        if get_aggregation_metrics(metric, max_depth):
            return max_depth
    return 0


def aggregate_metric_per_hierarchy_level(metric: dict[str, float | int], hierarchy_level: int) -> dict[str, float | int]:
    """Aggregates metrics by grouping and averaging values at a specified hierarchy level in the key structure.

    Args:
        metric: Dictionary mapping hierarchical keys (dot-separated strings) to numeric values.
               Array indices in keys are treated as wildcards.
        hierarchy_level: The depth level at which to aggregate the metrics.
                        E.g. level 1 aggregates at first dot separator.

    Returns:
        Dictionary mapping aggregated keys to averaged values. Keys are truncated to the specified
        hierarchy level with array indices replaced by '*'.

    Raises:
        ValueError: If the requested hierarchy level exceeds the maximum depth in the data.
    """
    max_depth = get_max_depth(metric)

    if hierarchy_level > max_depth:
        raise ValueError(f"Hierarchy level {hierarchy_level} is greater than the maximum depth {max_depth}")

    return get_aggregation_metrics(metric, hierarchy_level)



import hashlib
import hmac
import json
from typing import Any

from ...._resource import AsyncAPIResource, SyncAPIResource
from .endpoints import AsyncEndpoints, Endpoints
from .links import AsyncLinks, Links
from .mailboxes import AsyncMailboxes, Mailboxes
from .outlook import AsyncOutlooks, Outlooks
from .tests import AsyncTests, Tests
from .logs import AsyncLogs, Logs

class SignatureVerificationError(Exception):
    """Raised when webhook signature verification fails."""

    pass


class AutomationsMixin:
    def _verify_event(self, event_body: bytes, event_signature: str, secret: str) -> Any:
        """
        Verify the signature of a webhook event.

        Args:
            body: The raw request body
            signature: The signature header
            secret: The secret key used for signing

        Returns:
            The parsed event payload

        Raises:
            SignatureVerificationError: If the signature verification fails
        """
        expected_signature = hmac.new(secret.encode(), event_body, hashlib.sha256).hexdigest()

        if not hmac.compare_digest(event_signature, expected_signature):
            raise SignatureVerificationError("Invalid signature")

        return json.loads(event_body.decode('utf-8'))


class Automations(SyncAPIResource, AutomationsMixin):
    """Automations API wrapper"""

    def __init__(self, client: Any) -> None:
        super().__init__(client=client)
        self.mailboxes = Mailboxes(client=client)
        self.links = Links(client=client)
        self.outlook = Outlooks(client=client)
        self.endpoints = Endpoints(client=client)
        self.tests = Tests(client=client)
        self.logs = Logs(client=client)
    def verify_event(self, event_body: bytes, event_signature: str, secret: str) -> Any:
        """
        Verify the signature of a webhook event.
        """
        return self._verify_event(event_body, event_signature, secret)


class AsyncAutomations(AsyncAPIResource, AutomationsMixin):
    """Async Automations API wrapper"""

    def __init__(self, client: Any) -> None:
        super().__init__(client=client)
        self.mailboxes = AsyncMailboxes(client=client)
        self.links = AsyncLinks(client=client)
        self.outlook = AsyncOutlooks(client=client)
        self.endpoints = AsyncEndpoints(client=client)
        self.tests = AsyncTests(client=client)
        self.logs = AsyncLogs(client=client)

    async def verify_event(self, event_body: bytes, event_signature: str, secret: str) -> Any:
        """
        Verify the signature of a webhook event.
        """
        return self._verify_event(event_body, event_signature, secret)



import re
from typing import Any, Dict, List, Literal, Optional

import nanoid  # type: ignore
from pydantic import BaseModel, EmailStr, Field, field_validator, model_validator

from ..._utils.json_schema import convert_schema_to_layout
from ..logs import AutomationConfig, UpdateAutomationRequest
from ..pagination import ListMetadata

domain_pattern = re.compile(r"^(?:[a-zA-Z0-9-]+\.)+[a-zA-Z]{2,}$")


class AutomationLevel(BaseModel):
    distance_threshold: float = Field(default=0.9, description="Distance threshold for the automation")
    score_threshold: float = Field(default=0.9, description="Score threshold for the automation")


class MatchParams(BaseModel):
    endpoint: str = Field(..., description="Endpoint for matching parameters")
    headers: Dict[str, str] = Field(..., description="Headers for the request")
    path: str = Field(..., description="Path for matching parameters")


class FetchParams(BaseModel):
    endpoint: str = Field(..., description="Endpoint for fetching parameters")
    headers: Dict[str, str] = Field(..., description="Headers for the request")
    name: str = Field(..., description="Name of the fetch parameter")


class Outlook(AutomationConfig):
    object: Literal['automation.outlook'] = "automation.outlook"
    id: str = Field(default_factory=lambda: "outlook_" + nanoid.generate(), description="Unique identifier for the outlook")

    authorized_domains: list[str] = Field(default_factory=list, description="List of authorized domains to receive the emails from")
    authorized_emails: List[EmailStr] = Field(default_factory=list, description="List of emails to access the link")

    layout_schema: Optional[dict[str, Any]] = Field(default=None, description="Layout schema format used to display the data")

    # Optional Fields for data integration
    match_params: List[MatchParams] = Field(default_factory=list, description="List of match parameters for the outlook automation")
    fetch_params: List[FetchParams] = Field(default_factory=list, description="List of fetch parameters for the outlook automation")

    @model_validator(mode='before')
    @classmethod
    def compute_layout_schema(cls, values: dict[str, Any]) -> dict[str, Any]:
        if values.get('layout_schema') is None:
            values['layout_schema'] = convert_schema_to_layout(values['json_schema'])
        return values

class ListOutlooks(BaseModel):
    data: list[Outlook]
    list_metadata: ListMetadata


# Inherits from the methods of UpdateAutomationRequest
class UpdateOutlookRequest(UpdateAutomationRequest):
    authorized_domains: Optional[list[str]] = None
    authorized_emails: Optional[List[EmailStr]] = None

    match_params: Optional[List[MatchParams]] = None
    fetch_params: Optional[List[FetchParams]] = None

    layout_schema: Optional[dict[str, Any]] = None

    @field_validator("authorized_emails", mode="before")
    def normalize_authorized_emails(cls, emails: Optional[List[str]]) -> Optional[List[str]]:
        return [email.strip().lower() for email in emails] if emails else None


import base64
from typing import Any

import httpx
import nanoid  # type: ignore
import pytest
from pydantic import HttpUrl

from uiform import UiForm


@pytest.mark.asyncio
async def test_links_crud(sync_client: UiForm, company_json_schema: dict[str, Any], booking_confirmation_file_path: str) -> None:
    name = nanoid.generate()
    print("name", name)
    model = "gpt-4o-mini"
    webhook_url = HttpUrl('http://localhost:4000/v1/test_ingest_completion')
    # Create
    link = sync_client.processors.automations.links.create(
        name,
        company_json_schema,
        webhook_url,
        model=model,
    )
    link_id = link.id
    try:
        assert link.name == name
        assert link.webhook_url == webhook_url
        # Read
        link = sync_client.processors.automations.links.get(link_id)
        assert link.name == name
        # Update
        link = sync_client.processors.automations.links.update(link_id, password="password")
        assert link.password == "password"
        # Open and read the booking confirmation file

        # Let's upload a file
        with open(booking_confirmation_file_path, "rb") as f:
            async with httpx.AsyncClient(timeout=240) as client:
                usr_pwd_enc = base64.b64encode(f"{name}:password".encode("utf-8")).decode("utf-8")
                headers = {
                    "Authorization": f"Basic {usr_pwd_enc}",
                }
                files = {"file": f}
                response = await client.post(sync_client.base_url + f"/v1/processors/automations/links/parse/{link_id}", files=files, headers=headers)
                assert response.status_code == 200

        # Delete
        sync_client.processors.automations.links.delete(link_id)
        with pytest.raises(Exception):
            sync_client.processors.automations.links.get(link_id)
    finally:
        # Delete the link if it was created.
        try:
            sync_client.processors.automations.links.delete(link_id)
        except Exception:
            pass



import datetime
from typing import Optional

from openai.types.chat import completion_create_params
from openai.types.chat.chat_completion import ChatCompletion
from pydantic import BaseModel

from .._resource import AsyncAPIResource, SyncAPIResource
from ..types.ai_models import Amount
from ..types.logs import AutomationLog, LogCompletionRequest
from ..types.standards import PreparedRequest

total_cost = 0.0


class UsageMixin:
    def prepare_total(self, start_date: Optional[datetime.datetime] = None, end_date: Optional[datetime.datetime] = None) -> PreparedRequest:
        raise NotImplementedError("prepare_total is not implemented")

    def prepare_mailbox(self, email: str, start_date: Optional[datetime.datetime] = None, end_date: Optional[datetime.datetime] = None) -> PreparedRequest:
        params = {}
        if start_date:
            params["start_date"] = start_date.isoformat()
        if end_date:
            params["end_date"] = end_date.isoformat()

        return PreparedRequest(method="GET", url=f"/v1/processors/automations/mailboxes/{email}/usage", params=params)

    def prepare_link(self, link_id: str, start_date: Optional[datetime.datetime] = None, end_date: Optional[datetime.datetime] = None) -> PreparedRequest:
        params = {}
        if start_date:
            params["start_date"] = start_date.isoformat()
        if end_date:
            params["end_date"] = end_date.isoformat()

        return PreparedRequest(method="GET", url=f"/v1/processors/automations/links/{link_id}/usage", params=params)

    def prepare_schema(self, schema_id: str, start_date: Optional[datetime.datetime] = None, end_date: Optional[datetime.datetime] = None) -> PreparedRequest:
        params = {}
        if start_date:
            params["start_date"] = start_date.isoformat()
        if end_date:
            params["end_date"] = end_date.isoformat()

        return PreparedRequest(method="GET", url=f"/v1/schemas/{schema_id}/usage", params=params)

    def prepare_schema_data(self, schema_data_id: str, start_date: Optional[datetime.datetime] = None, end_date: Optional[datetime.datetime] = None) -> PreparedRequest:
        params = {}
        if start_date:
            params["start_date"] = start_date.isoformat()
        if end_date:
            params["end_date"] = end_date.isoformat()

        return PreparedRequest(method="GET", url=f"/v1/schemas/{schema_data_id}/usage_data", params=params)

    def prepare_log(self, response_format: completion_create_params.ResponseFormat, completion: ChatCompletion) -> PreparedRequest:
        if isinstance(response_format, BaseModel):
            log_completion_request = LogCompletionRequest(json_schema=response_format.model_json_schema(), completion=completion)
        elif isinstance(response_format, dict):
            if "json_schema" in response_format:
                json_schema = response_format["json_schema"]  # type: ignore
                if "schema" in json_schema:
                    log_completion_request = LogCompletionRequest(json_schema=json_schema["schema"], completion=completion)
                else:
                    raise ValueError("Invalid response format")
            else:
                raise ValueError("Invalid response format")
        else:
            raise ValueError("Invalid response format")

        return PreparedRequest(method="POST", url="/v1/usage/log", data=log_completion_request.model_dump())


class Usage(SyncAPIResource, UsageMixin):
    def total(self, start_date: Optional[datetime.datetime] = None, end_date: Optional[datetime.datetime] = None) -> Amount:
        """Get the total usage cost for a mailbox within an optional date range.

        Returns:
            Amount: The total usage cost
        """
        return Amount(value=total_cost, currency="USD")

    def mailbox(self, email: str, start_date: Optional[datetime.datetime] = None, end_date: Optional[datetime.datetime] = None) -> Amount:
        """Get the total usage cost for a mailbox within an optional date range.

        Args:
            email: The email address of the mailbox
            start_date: Start date for usage calculation
            end_date: End date for usage calculation

        Returns:
            Amount: The total usage cost
        """
        request = self.prepare_mailbox(email, start_date, end_date)
        response = self._client._request(request.method, request.url, request.data, request.params)
        return Amount.model_validate(response)

    def link(self, link_id: str, start_date: Optional[datetime.datetime] = None, end_date: Optional[datetime.datetime] = None) -> Amount:
        """Get the total usage cost for a link within an optional date range.

        Args:
            link_id: The ID of the link
            start_date: Start date for usage calculation
            end_date: End date for usage calculation

        Returns:
            Amount: The total usage cost
        """
        request = self.prepare_link(link_id, start_date, end_date)
        response = self._client._request(request.method, request.url, request.data, request.params)
        return Amount.model_validate(response)

    def schema(self, schema_id: str, start_date: Optional[datetime.datetime] = None, end_date: Optional[datetime.datetime] = None) -> Amount:
        """Get the total usage cost for a schema within an optional date range.

        Args:
            schema_id: The ID of the schema
            start_date: Start date for usage calculation
            end_date: End date for usage calculation

        Returns:
            Amount: The total usage cost
        """
        request = self.prepare_schema(schema_id, start_date, end_date)
        response = self._client._request(request.method, request.url, request.data, request.params)
        return Amount.model_validate(response)

    def schema_data(self, schema_data_id: str, start_date: Optional[datetime.datetime] = None, end_date: Optional[datetime.datetime] = None) -> Amount:
        """Get the total usage cost for a schema within an optional date range.

        Args:
            schema_id: The ID of the schema
            start_date: Start date for usage calculation
            end_date: End date for usage calculation

        Returns:
            Amount: The total usage cost
        """
        request = self.prepare_schema_data(schema_data_id, start_date, end_date)
        response = self._client._request(request.method, request.url, request.data, request.params)
        return Amount.model_validate(response)

    # TODO: Turn that into an async process
    def log(self, response_format: completion_create_params.ResponseFormat, completion: ChatCompletion) -> AutomationLog:
        """Logs an openai request completion as an automation log to make the usage calculation possible for the user

        client = OpenAI()
        completion = client.beta.chat.completions.parse(
            model="gpt-4o-2024-08-06",
            messages=[
                {"role": "developer", "content": "Extract the event information."},
                {"role": "user", "content": "Alice and Bob are going to a science fair on Friday."},
            ],
            response_format=CalendarEvent,
        )
        uiclient.usage.log(
            response_format=CalendarEvent,
            completion=completion
        )


        Args:
            response_format: The response format of the openai request
            completion: The completion of the openai request

        Returns:
            AutomationLog: The automation log
        """
        request = self.prepare_log(response_format, completion)
        response = self._client._request(request.method, request.url, request.data, request.params)
        return AutomationLog.model_validate(response)


class AsyncUsage(AsyncAPIResource, UsageMixin):
    async def total(self, start_date: Optional[datetime.datetime] = None, end_date: Optional[datetime.datetime] = None) -> Amount:
        """Get the total usage cost for a mailbox within an optional date range.

        Returns:
            Amount: The total usage cost
        """
        return Amount(value=total_cost, currency="USD")

    async def mailbox(self, email: str, start_date: Optional[datetime.datetime] = None, end_date: Optional[datetime.datetime] = None) -> Amount:
        """Get the total usage cost for a mailbox within an optional date range.

        Args:
            email: The email address of the mailbox
            start_date: Start date for usage calculation
            end_date: End date for usage calculation

        Returns:
            Amount: The total usage cost
        """
        request = self.prepare_mailbox(email, start_date, end_date)
        response = await self._client._request(request.method, request.url, request.data, request.params)
        return Amount.model_validate(response)

    async def link(self, link_id: str, start_date: Optional[datetime.datetime] = None, end_date: Optional[datetime.datetime] = None) -> Amount:
        """Get the total usage cost for a link within an optional date range.

        Args:
            link_id: The ID of the link
            start_date: Start date for usage calculation
            end_date: End date for usage calculation

        Returns:
            Amount: The total usage cost
        """
        request = self.prepare_link(link_id, start_date, end_date)
        response = await self._client._request(request.method, request.url, request.data, request.params)
        return Amount.model_validate(response)

    async def schema(self, schema_id: str, start_date: Optional[datetime.datetime] = None, end_date: Optional[datetime.datetime] = None) -> Amount:
        """Get the total usage cost for a schema within an optional date range.

        Args:
            schema_id: The ID of the schema
            start_date: Start date for usage calculation
            end_date: End date for usage calculation

        Returns:
            Amount: The total usage cost
        """
        request = self.prepare_schema(schema_id, start_date, end_date)
        response = await self._client._request(request.method, request.url, request.data, request.params)
        return Amount.model_validate(response)

    async def schema_data(self, schema_data_id: str, start_date: Optional[datetime.datetime] = None, end_date: Optional[datetime.datetime] = None) -> Amount:
        """Get the total usage cost for a schema within an optional date range.

        Args:
            schema_id: The ID of the schema
            start_date: Start date for usage calculation
            end_date: End date for usage calculation

        Returns:
            Amount: The total usage cost
        """
        request = self.prepare_schema_data(schema_data_id, start_date, end_date)
        response = await self._client._request(request.method, request.url, request.data, request.params)
        return Amount.model_validate(response)

    # TODO: Turn that into an async process
    async def log(self, response_format: completion_create_params.ResponseFormat, completion: ChatCompletion) -> AutomationLog:
        """Logs an openai request completion as an automation log to make the usage calculation possible for the user

        client = OpenAI()
        completion = client.beta.chat.completions.parse(
            model="gpt-4o-2024-08-06",
            messages=[
                {"role": "developer", "content": "Extract the event information."},
                {"role": "user", "content": "Alice and Bob are going to a science fair on Friday."},
            ],
            response_format=CalendarEvent,
        )
        uiclient.usage.log(
            response_format=CalendarEvent,
            completion=completion
        )


        Args:
            response_format: The response format of the openai request
            completion: The completion of the openai request

        Returns:
            AutomationLog: The automation log
        """
        request = self.prepare_log(response_format, completion)
        response = await self._client._request(request.method, request.url, request.data, request.params)
        return AutomationLog.model_validate(response)



import asyncio
import json
import time
from typing import Any, Literal, get_args

import httpx
import nanoid  # type: ignore
import pytest
from pydantic import BaseModel

from uiform import AsyncUiForm, UiForm
from uiform.types.documents.extractions import UiParsedChatCompletion

# List of AI Providers to test
AI_MODELS = Literal[
    "gpt-4o-mini",
    # "grok-2-vision-1212",
    # "claude-3-5-sonnet-latest",
    # "gemini-1.5-flash-8b"
]
ClientType = Literal[
    "sync",
    "async",
]
ResponseModeType = Literal[
    "stream",
    "parse",
]


def validate_extraction_response(response: UiParsedChatCompletion | None) -> None:
    # Assert the instance
    assert isinstance(response, UiParsedChatCompletion), f"Response should be of type UiParsedChatCompletion, received {type(response)}"

    # Assert the response content is not None
    assert response.choices[0].message.content is not None, "Response content should not be None"
    # Assert that content is a valid JSON object
    try:
        json.loads(response.choices[0].message.content)
    except json.JSONDecodeError:
        assert False, "Response content should be a valid JSON object"
    # Assert that the response.choices[0].message.parsed is a valid pydantic BaseModel instance
    assert isinstance(response.choices[0].message.parsed, BaseModel), "Response parsed should be a valid pydantic BaseModel instance"


# Test the extraction endpoint
async def base_test_extract(
    model: AI_MODELS,
    client_type: ClientType,
    response_mode: ResponseModeType,
    sync_client: UiForm,
    async_client: AsyncUiForm,
    booking_confirmation_file_path: str,
    booking_confirmation_json_schema: dict[str, Any],
) -> None:
    json_schema = booking_confirmation_json_schema
    document = booking_confirmation_file_path
    modality: Literal["text"] = "text"
    response: UiParsedChatCompletion | None = None
    # Wait a random amount of time between 0 and 2 seconds (to avoid sending multiple requests to the same provider at the same time)
    if client_type == "sync":
        with sync_client as client:
            if response_mode == "stream":
                with client.documents.extractions.stream(json_schema=json_schema, document=document, model=model, modality=modality) as stream_iterator:
                    response = stream_iterator.__next__()
                    for response in stream_iterator:
                        pass
            else:
                response = client.documents.extractions.parse(json_schema=json_schema, document=document, model=model, modality=modality)
    if client_type == "async":
        async with async_client:
            if response_mode == "stream":
                async with async_client.documents.extractions.stream(json_schema=json_schema, document=document, model=model, modality=modality) as stream_async_iterator:
                    response = await stream_async_iterator.__anext__()
                    async for response in stream_async_iterator:
                        pass
            else:
                response = await async_client.documents.extractions.parse(json_schema=json_schema, document=document, model=model, modality=modality)
    validate_extraction_response(response)


@pytest.mark.asyncio
@pytest.mark.xdist_group(name="openai")
@pytest.mark.parametrize("client_type", get_args(ClientType))
@pytest.mark.parametrize("response_mode", get_args(ResponseModeType))
async def test_extract_openai(
    client_type: ClientType,
    response_mode: ResponseModeType,
    sync_client: UiForm,
    async_client: AsyncUiForm,
    booking_confirmation_file_path: str,
    booking_confirmation_json_schema: dict[str, Any],
) -> None:
    await base_test_extract(
        model="gpt-4o-mini",
        client_type=client_type,
        response_mode=response_mode,
        sync_client=sync_client,
        async_client=async_client,
        booking_confirmation_file_path=booking_confirmation_file_path,
        booking_confirmation_json_schema=booking_confirmation_json_schema,
    )


@pytest.mark.asyncio
@pytest.mark.parametrize("request_number", range(10))
async def test_extract_overload(
    sync_client: UiForm,
    async_client: AsyncUiForm,
    booking_confirmation_file_path: str,
    booking_confirmation_json_schema: dict[str, Any],
    request_number: int,
) -> None:
    await asyncio.sleep(request_number * 0.1)
    await base_test_extract(
        model="gpt-4o-mini",
        client_type="async",
        response_mode="parse",
        sync_client=sync_client,
        async_client=async_client,
        booking_confirmation_file_path=booking_confirmation_file_path,
        booking_confirmation_json_schema=booking_confirmation_json_schema,
    )


# @pytest.mark.asyncio
# @pytest.mark.xdist_group(name="anthropic")
# @pytest.mark.parametrize("client_type", get_args(ClientType))
# @pytest.mark.parametrize("response_mode", get_args(ResponseModeType))
# async def test_extract_anthropic(client_type: ClientType, response_mode: ResponseModeType, sync_client: UiForm, async_client: AsyncUiForm, booking_confirmation_file_path: str, booking_confirmation_json_schema: dict[str, Any]) -> None:
#     await base_test_extract(
#         model="claude-3-5-sonnet-latest",
#         client_type=client_type,
#         response_mode=response_mode,
#         sync_client=sync_client,
#         async_client=async_client,
#         booking_confirmation_file_path=booking_confirmation_file_path,
#         booking_confirmation_json_schema=booking_confirmation_json_schema
#     )

# @pytest.mark.asyncio
# @pytest.mark.xdist_group(name="xai")
# @pytest.mark.parametrize("client_type", get_args(ClientType))
# @pytest.mark.parametrize("response_mode", get_args(ResponseModeType))
# async def test_extract_xai(client_type: ClientType, response_mode: ResponseModeType, sync_client: UiForm, async_client: AsyncUiForm, booking_confirmation_file_path: str, booking_confirmation_json_schema: dict[str, Any]) -> None:
#     await base_test_extract(
#         model="grok-2-vision-1212",
#         client_type=client_type,
#         response_mode=response_mode,
#         sync_client=sync_client,
#         async_client=async_client,
#         booking_confirmation_file_path=booking_confirmation_file_path,
#         booking_confirmation_json_schema=booking_confirmation_json_schema
#     )

# @pytest.mark.asyncio
# @pytest.mark.xdist_group(name="google")
# @pytest.mark.parametrize("client_type", get_args(ClientType))
# @pytest.mark.parametrize("response_mode", get_args(ResponseModeType))
# async def test_extract_google(client_type: ClientType, response_mode: ResponseModeType, sync_client: UiForm, async_client: AsyncUiForm, booking_confirmation_file_path: str, booking_confirmation_json_schema: dict[str, Any]) -> None:
#     await base_test_extract(
#         model="gemini-1.5-flash-8b",
#         client_type=client_type,
#         response_mode=response_mode,
#         sync_client=sync_client,
#         async_client=async_client,
#         booking_confirmation_file_path=booking_confirmation_file_path,
#         booking_confirmation_json_schema=booking_confirmation_json_schema
#     )


@pytest.mark.asyncio
async def test_extraction_with_idempotency(sync_client: UiForm, booking_confirmation_file_path: str, booking_confirmation_json_schema: dict[str, Any]) -> None:
    client = sync_client
    idempotency_key = nanoid.generate()
    response_initial = client.documents.extractions.parse(
        json_schema=booking_confirmation_json_schema, document=booking_confirmation_file_path, model="gpt-4o-mini", modality="native", idempotency_key=idempotency_key
    )
    await asyncio.sleep(2)
    t0 = time.time()
    response_second = client.documents.extractions.parse(
        json_schema=booking_confirmation_json_schema, document=booking_confirmation_file_path, model="gpt-4o-mini", modality="native", idempotency_key=idempotency_key
    )
    t1 = time.time()
    assert t1 - t0 < 10, "Request should take less than 10 seconds"
    assert response_initial.choices[0].message.content == response_second.choices[0].message.content, "Response should be the same"
    assert response_initial.choices[0].message.parsed is not None, "Parsed response should not be None for the first request"
    assert response_second.choices[0].message.parsed is not None, "Parsed response should not be None for the second request"
    assert response_initial.choices[0].message.parsed.model_dump() == response_second.choices[0].message.parsed.model_dump(), "Parsed response should be the same"


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "test_exception",
    ["before_handle_extraction", "within_extraction_parse_or_stream", "after_handle_extraction", "within_process_document_stream_generator"],
)
async def test_extraction_with_idempotency_exceptions(
    sync_client: UiForm, booking_confirmation_file_path: str, booking_confirmation_json_schema: dict[str, Any], test_exception: str
) -> None:
    client = sync_client
    # Now we will validate some exception scenarios
    idempotency_key = str(nanoid.generate())
    loading_request = client.documents.extractions.prepare_extraction(
        json_schema=booking_confirmation_json_schema,
        document=booking_confirmation_file_path,
        image_resolution_dpi=96,
        browser_canvas="A4",
        model="gpt-4o-mini",
        temperature=0,
        modality="native",
        reasoning_effort="medium",
        stream=False,
        n_consensus=1,
        store=False,
        idempotency_key=idempotency_key,
    )
    assert loading_request.data is not None, "Loading request should not be None"

    # Validate idempotency behavior for each of the following scenarios:
    loading_request.data["test_exception"] = test_exception
    response1: Any = None
    response2: Any = None
    raised_exception_1: Exception | None = None
    raised_exception_2: Exception | None = None
    try:
        loading_request.raise_for_status = True
        if test_exception == "within_process_document_stream_generator":
            loading_request.data["stream"] = True
            stream_iterator = client._prepared_request_stream(loading_request)
            response1 = stream_iterator.__next__()
            for response1 in stream_iterator:
                pass
        else:
            response1 = client._prepared_request(loading_request)
    except Exception as e:
        raised_exception_1 = e

    await asyncio.sleep(2)
    t0 = time.time()
    try:
        loading_request.raise_for_status = True
        if test_exception == "within_process_document_stream_generator":
            loading_request.data["stream"] = True
            stream_iterator = client._prepared_request_stream(loading_request)
            response2 = stream_iterator.__next__()
            for response2 in stream_iterator:
                pass
        else:
            response2 = client._prepared_request(loading_request)
    except Exception as e:
        raised_exception_2 = e
    t1 = time.time()
    assert t1 - t0 < 10, "Request should take less than 10 seconds"
    assert raised_exception_1 is not None, "Exception should be raised"
    assert raised_exception_2 is not None, "Exception should be raised"
    assert response1 is None, "Response should be None"
    assert response2 is None, "Response should be None"
    # Assert that the both exceptions is a HTTPStatusError
    assert isinstance(raised_exception_1, httpx.HTTPStatusError), "Exception should be a HTTPStatusError"
    assert isinstance(raised_exception_2, httpx.HTTPStatusError), "Exception should be a HTTPStatusError"
    # Assert that the message of both exceptions is the same
    assert raised_exception_1.args[0] == raised_exception_2.args[0], "Exception message should be the same"



import datetime
from typing import Any, Literal, Optional

import nanoid  # type: ignore
from pydantic import BaseModel, Field

metadata_key = Literal[
    'user',
    'organization',
    'link',
    'mailbox',
    'cron',
    'outlook',
    'extraction',
    'webhook',
    'reconciliation',
    'preprocessing',
    'schema',
    'data_structure',
    'file',
    'preprocessing',
    'dataset',
    'dataset_membership',
    'endpoint',
    'automation',
    'template',
]

event_type = Literal[
    'extraction.created',
    'messages.created',
    'document.orientation_corrected',
    'consensus.reconciled',
    'automation.created',
    'automation.updated',
    'automation.deleted',
    'automation.webhook',
    'preprocessing.created',
    'link.created',
    'link.updated',
    'link.deleted',
    'link.webhook',
    'mailbox.created',
    'mailbox.updated',
    'mailbox.deleted',
    'mailbox.webhook',
    'outlook.created',
    'outlook.updated',
    'outlook.deleted',
    'outlook.webhook',
    'schema.generated',
    'schema.promptified',
    'schema.system_promptfile.created',
    'file.updated',
    'file.deleted',
    'template.created',
    'template.deleted',
    'template.sample_document_uploaded',
    'template.sample_document_deleted',
    'template.updated',
]


class Event(BaseModel):
    object: Literal['event'] = "event"
    id: str = Field(default_factory=lambda: "event_" + nanoid.generate(), description="Unique identifier for the event")
    event: str = Field(..., description="A string that distinguishes the event type. Ex: user.created, user.updated, user.deleted, etc.")
    created_at: datetime.datetime = Field(default_factory=lambda: datetime.datetime.now(datetime.timezone.utc))
    data: dict[str, Any] = Field(..., description="Event payload. Payloads match the corresponding API objects.")
    metadata: Optional[dict[metadata_key, str]] = Field(
        default=None, description="Ids giving informations about the event. Ex: user.created.metadata = {'user': 'usr_8478973619047837'}"
    )


class StoredEvent(Event):
    organization_id: str = Field(..., description="Organization ID")



from openai.types.model import Model

from .._resource import AsyncAPIResource, SyncAPIResource
from ..types.standards import PreparedRequest


class ModelsMixin:
    def prepare_list(self) -> PreparedRequest:
        return PreparedRequest(method="GET", url="/v1/models")


class Models(SyncAPIResource, ModelsMixin):
    """Models API wrapper"""

    def list(self) -> list[Model]:
        """
        List all available models.

        Returns:
            list[str]: List of available models
        Raises:
            HTTPException if the request fails
        """

        request = self.prepare_list()
        output = self._client._prepared_request(request)
        return [Model(**model) for model in output["data"]]


class AsyncModels(AsyncAPIResource, ModelsMixin):
    """Models Asyncronous API wrapper"""

    async def list(self) -> list[Model]:
        """
        List all available models.

        Returns:
            list[str]: List of available models
        Raises:
            HTTPException if the request fails
        """

        request = self.prepare_list()
        output = await self._client._prepared_request(request)
        return [Model(**model) for model in output["data"]]



from pydantic import BaseModel, Field, model_validator
from typing import Any, Literal, Self
from ..mime import MIMEData
from ..modalities import Modality
from openai.types.chat.chat_completion_reasoning_effort import ChatCompletionReasoningEffort
from ..evals import ItemMetric


class EvaluateSchemaRequest(BaseModel):
    """
    The request body for evaluating a JSON Schema.

    Two evaluation modes are supported:
    - 'consensus': Uses multiple extractions to compute consensus-based likelihoods
    - 'distance': Compares single extraction against provided ground truths
    """

    documents: list[MIMEData]
    ground_truths: list[dict[str, Any]] | None = None
    model: str = "gpt-4o-mini"
    reasoning_effort: ChatCompletionReasoningEffort = "medium"
    modality: Modality
    image_resolution_dpi: int = Field(default=96, description="Resolution of the image sent to the LLM")
    browser_canvas: Literal["A3", "A4", "A5"] = Field(
        default="A4", description="Sets the size of the browser canvas for rendering documents in browser-based processing. Choose a size that matches the document type."
    )
    n_consensus: int = 1
    json_schema: dict[str, Any]

    @model_validator(mode="after")
    def validate_evaluation_mode_consistency(self) -> Self:
        """Validate that parameters are consistent with the chosen evaluation mode."""

        if self.ground_truths is None:
            # Consensus mode requirements
            if self.n_consensus <= 1:
                raise ValueError("Consensus mode requires n_consensus > 1")
        else:
            # Distance mode requirements
            if self.n_consensus != 1:
                raise ValueError("Distance mode requires n_consensus = 1")
            if len(self.documents) != len(self.ground_truths):
                raise ValueError("Distance mode requires equal number of documents and ground_truths")
        if len(self.documents) == 0:
            raise ValueError("Evaluation mode requires at least one document")

        return self


class EvaluateSchemaResponse(BaseModel):
    """
    The response body for evaluating a JSON Schema.
    """

    item_metrics: list[ItemMetric]



from typing import Literal

import nanoid  # type: ignore
from pydantic import BaseModel, Field

from ..logs import AutomationConfig, UpdateAutomationRequest
from ..pagination import ListMetadata


class Endpoint(AutomationConfig):
    object: Literal['automation.endpoint'] = "automation.endpoint"
    id: str = Field(default_factory=lambda: "endp_" + nanoid.generate(), description="Unique identifier for the extraction endpoint")

class ListEndpoints(BaseModel):
    data: list[Endpoint]
    list_metadata: ListMetadata


# Inherits from the methods of UpdateAutomationRequest
class UpdateEndpointRequest(UpdateAutomationRequest):
    pass



import datetime
import json
from io import IOBase
from pathlib import Path
from typing import Any, Dict, List, Literal, Optional

import httpx
from openai.types.chat.chat_completion_reasoning_effort import ChatCompletionReasoningEffort
from PIL.Image import Image
from pydantic import HttpUrl

from ...._resource import AsyncAPIResource, SyncAPIResource
from ...._utils.ai_models import assert_valid_model_extraction
from ...._utils.mime import prepare_mime_document
from ....types.automations.mailboxes import ListMailboxes, Mailbox, UpdateMailboxRequest
from ....types.documents.extractions import UiParsedChatCompletion
from ....types.logs import AutomationLog, ExternalRequestLog
from ....types.mime import BaseMIMEData, EmailData, MIMEData
from ....types.modalities import Modality
from ....types.standards import PreparedRequest


class MailBoxesMixin:
    def prepare_create(
        self,
        email: str,
        name: str,
        json_schema: Dict[str, Any],
        webhook_url: HttpUrl,
        # email specific opitonals Fields
        authorized_domains: List[str] = [],
        authorized_emails: List[str] = [],
        # HTTP Config Optional Fields
        webhook_headers: Dict[str, str] = {},
        # DocumentExtraction Config
        image_resolution_dpi: int = 96,
        browser_canvas: Literal['A3', 'A4', 'A5'] = 'A4',
        modality: Modality = "native",
        model: str = "gpt-4o-mini",
        temperature: float = 0,
        reasoning_effort: ChatCompletionReasoningEffort = "medium",
    ) -> PreparedRequest:
        assert_valid_model_extraction(model)

        data = {
            "email": email,
            "name": name,
            "webhook_url": webhook_url,
            "webhook_headers": webhook_headers,
            "json_schema": json_schema, 
            "authorized_domains": authorized_domains,
            "authorized_emails": authorized_emails,
            "image_resolution_dpi": image_resolution_dpi,
            "browser_canvas": browser_canvas,
            "modality": modality,
            "model": model,
            "temperature": temperature,
            "reasoning_effort": reasoning_effort,
        }

        # Validate the data
        mailbox_data = Mailbox.model_validate(data)
        return PreparedRequest(method="POST", url="/v1/deployments/mailboxes", data=mailbox_data.model_dump(mode="json"))

    def prepare_list(
        self,
        before: str | None = None,
        after: str | None = None,
        limit: int = 10,
        order: Literal["asc", "desc"] | None = "desc",
        email: Optional[str] = None,
        name: Optional[str] = None,
        webhook_url: Optional[str] = None,
        schema_id: Optional[str] = None,
        schema_data_id: Optional[str] = None,
    ) -> PreparedRequest:
        params = {
            "before": before,
            "after": after,
            "limit": limit,
            "order": order,
            "email": email,
            "name": name,
            "webhook_url": webhook_url,
            "schema_id": schema_id,
            "schema_data_id": schema_data_id,
        }
        # Remove None values
        params = {k: v for k, v in params.items() if v is not None}

        return PreparedRequest(method="GET", url="/v1/deployments/mailboxes", params=params)

    def prepare_get(self, email: str) -> PreparedRequest:
        return PreparedRequest(method="GET", url=f"/v1/processors/automations/mailboxes/{email}")

    def prepare_update(
        self,
        email: str,
        name: Optional[str] = None,
        webhook_url: Optional[HttpUrl] = None,
        webhook_headers: Optional[Dict[str, str]] = None,
        authorized_domains: Optional[List[str]] = None,
        authorized_emails: Optional[List[str]] = None,
        image_resolution_dpi: Optional[int] = None,
        browser_canvas: Optional[Literal['A3', 'A4', 'A5']] = None,
        modality: Optional[Modality] = None,
        model: Optional[str] = None,
        temperature: Optional[float] = None,
        reasoning_effort: Optional[ChatCompletionReasoningEffort] = None,
        json_schema: Optional[Dict[str, Any]] = None,
    ) -> PreparedRequest:
        # Build data dict excluding None values
        data = {
            k: v for k, v in {
                "name": name,
                "webhook_url": webhook_url,
                "webhook_headers": webhook_headers,
                "authorized_domains": authorized_domains,
                "authorized_emails": authorized_emails,
                "image_resolution_dpi": image_resolution_dpi,
                "browser_canvas": browser_canvas,
                "modality": modality,
                "model": model,
                "temperature": temperature,
                "reasoning_effort": reasoning_effort,
                "json_schema": json_schema,
            }.items() if v is not None
        }

        if "model" in data:
            assert_valid_model_extraction(data["model"])

        update_mailbox_request = UpdateMailboxRequest.model_validate(data)
        return PreparedRequest(
            method="PUT", 
            url=f"/v1/processors/automations/mailboxes/{email}", 
            data=update_mailbox_request.model_dump(mode="json")
        )

    def prepare_delete(self, email: str) -> PreparedRequest:
        return PreparedRequest(method="DELETE", url=f"/v1/processors/automations/mailboxes/{email}", raise_for_status=True)

class Mailboxes(SyncAPIResource, MailBoxesMixin):
    """Emails API wrapper for managing email automation configurations"""

    def __init__(self, client: Any) -> None:
        super().__init__(client=client)
        self.tests = TestMailboxes(client=client)

    def create(
        self,
        email: str,
        name: str,
        json_schema: Dict[str, Any],
        webhook_url: HttpUrl,
        # email specific opitonals Fields
        authorized_domains: List[str] = [],
        authorized_emails: List[str] = [],
        # HTTP Config Optional Fields
        webhook_headers: Dict[str, str] = {},
        # DocumentExtraction Config
        image_resolution_dpi: int = 96,
        browser_canvas: Literal['A3', 'A4', 'A5'] = 'A4',
        modality: Modality = "native",
        model: str = "gpt-4o-mini",
        temperature: float = 0,
        reasoning_effort: ChatCompletionReasoningEffort = "medium",
    ) -> Mailbox:
        """Create a new email automation configuration.

        Args:
            email: Email address for the mailbox
            json_schema: JSON schema to validate extracted email data
            webhook_url: Webhook URL to receive processed emails
            webhook_headers: Webhook headers to send with processed emails
            authorized_domains: List of authorized domains for the mailbox
            authorized_emails: List of authorized emails for the mailbox
            image_resolution_dpi: Image resolution DPI
            browser_canvas: Browser canvas size
            modality: Processing modality (currently only "native" supported)
            model: AI model to use for processing
            temperature: Model temperature setting
            reasoning_effort: The effort level for the model to reason about the input data.

        Returns:
            Mailbox: The created mailbox configuration
        """

        request = self.prepare_create(
            email, name, json_schema, webhook_url, authorized_domains, authorized_emails, webhook_headers, image_resolution_dpi, browser_canvas, modality, model, temperature, reasoning_effort
        )
        response = self._client._prepared_request(request)

        print(f"Email automation created. Mailbox available at https://www.uiform.com/dashboard/processors/{response['id']}")

        return Mailbox.model_validate(response)

    def list(
        self,
        before: str | None = None,
        after: str | None = None,
        limit: int = 10,
        order: Literal["asc", "desc"] | None = "desc",
        email: Optional[str] = None,
        webhook_url: Optional[str] = None,
        schema_id: Optional[str] = None,
        schema_data_id: Optional[str] = None,
    ) -> ListMailboxes:
        """List all email automation configurations.

        Args:
            before: Optional cursor for pagination - get results before this log ID
            after: Optional cursor for pagination - get results after this log ID
            limit: Maximum number of logs to return (1-100, default 10)
            order: Sort order by creation time - "asc" or "desc" (default "desc")
            email: Optional email address filter
            webhook_url: Optional webhook URL filter
            schema_id: Optional schema ID filter
            schema_data_id: Optional schema data ID filter

        Returns:
            ListMailboxes: List of mailbox configurations
        """
        request = self.prepare_list(before, after, limit, order, email, webhook_url, schema_id, schema_data_id)
        response = self._client._prepared_request(request)
        return ListMailboxes.model_validate(response)

    def get(self, email: str) -> Mailbox:
        """Get a specific email automation configuration.

        Args:
            email: Email address of the mailbox

        Returns:
            Mailbox: The mailbox configuration
        """
        request = self.prepare_get(email)
        response = self._client._prepared_request(request)
        return Mailbox.model_validate(response)

    def update(
        self,
        email: str,
        name: Optional[str] = None,
        webhook_url: Optional[HttpUrl] = None,
        webhook_headers: Optional[Dict[str, str]] = None,
        authorized_domains: Optional[List[str]] = None,
        authorized_emails: Optional[List[str]] = None,
        image_resolution_dpi: Optional[int] = None,
        browser_canvas: Optional[Literal['A3', 'A4', 'A5']] = None,
        modality: Optional[Modality] = None,
        model: Optional[str] = None,
        temperature: Optional[float] = None,
        reasoning_effort: Optional[ChatCompletionReasoningEffort] = None,
        json_schema: Optional[Dict[str, Any]] = None,
    ) -> Mailbox:
        """Update an email automation configuration.

        Args:
            email: Email address of the mailbox to update
            webhook_url: New webhook configuration
            webhook_headers: New webhook configuration
            max_file_size: New webhook configuration
            file_payload: New webhook configuration
            follow_up: New webhook configuration
            authorized_domains: New webhook configuration
            authorized_emails: New webhook configuration
            image_resolution_dpi: New image resolution DPI
            browser_canvas: New browser canvas size
            modality: New processing modality
            model: New AI model
            temperature: New temperature setting
            reasoning_effort: New reasoning effort
            json_schema: New JSON schema

        Returns:
            Mailbox: The updated mailbox configuration
        """
        request = self.prepare_update(
            email, name, webhook_url, webhook_headers, authorized_domains, authorized_emails, image_resolution_dpi, browser_canvas, modality, model, temperature, reasoning_effort, json_schema
        )
        response = self._client._prepared_request(request)
        return Mailbox.model_validate(response)

    def delete(self, email: str) -> None:
        """Delete an email automation configuration.

        Args:
            email: Email address of the mailbox to delete
        """
        request = self.prepare_delete(email)
        response = self._client._prepared_request(request)
        return None

class AsyncMailboxes(AsyncAPIResource, MailBoxesMixin):
    def __init__(self, client: Any) -> None:
        super().__init__(client=client)
        self.tests = AsyncTestMailboxes(client=client)

    async def create(
        self,
        email: str,
        name: str,
        json_schema: Dict[str, Any],
        webhook_url: HttpUrl,
        authorized_domains: List[str] = [],
        authorized_emails: List[str] = [],
        webhook_headers: Dict[str, str] = {},
        image_resolution_dpi: int = 96,
        browser_canvas: Literal['A3', 'A4', 'A5'] = 'A4',
        modality: Modality = "native",
        model: str = "gpt-4o-mini",
        temperature: float = 0,
        reasoning_effort: ChatCompletionReasoningEffort = "medium",
    ) -> Mailbox:
        request = self.prepare_create(
                email, name, json_schema, webhook_url, authorized_domains, authorized_emails, webhook_headers, image_resolution_dpi, browser_canvas, modality, model, temperature, reasoning_effort
        )
        response = await self._client._prepared_request(request)

        print(f"Email automation created. Mailbox available at https://www.uiform.com/dashboard/processors/{response['id']}")

        return Mailbox.model_validate(response)

    async def list(
        self,
        before: str | None = None,
        after: str | None = None,
        limit: int = 10,
        order: Literal["asc", "desc"] | None = "desc",
        email: Optional[str] = None,
        name: Optional[str] = None,
        webhook_url: Optional[str] = None,
        schema_id: Optional[str] = None,
        schema_data_id: Optional[str] = None,
    ) -> ListMailboxes:
        request = self.prepare_list(before, after, limit, order, email, webhook_url, schema_id, schema_data_id)
        response = await self._client._prepared_request(request)
        return ListMailboxes.model_validate(response)

    async def get(self, email: str) -> Mailbox:
        request = self.prepare_get(email)
        response = await self._client._prepared_request(request)
        return Mailbox.model_validate(response)

    async def update(
        self,
        email: str,
        name: Optional[str] = None,
        webhook_url: Optional[HttpUrl] = None,
        webhook_headers: Optional[Dict[str, str]] = None,
        authorized_domains: Optional[List[str]] = None,
        authorized_emails: Optional[List[str]] = None,
        image_resolution_dpi: Optional[int] = None,
        browser_canvas: Optional[Literal['A3', 'A4', 'A5']] = None,
        modality: Optional[Modality] = None,
        model: Optional[str] = None,
        temperature: Optional[float] = None,
        reasoning_effort: Optional[ChatCompletionReasoningEffort] = None,
        json_schema: Optional[Dict[str, Any]] = None,
    ) -> Mailbox:
        request = self.prepare_update(
            email, name, webhook_url, webhook_headers, authorized_domains, authorized_emails, image_resolution_dpi, browser_canvas, modality, model, temperature, reasoning_effort, json_schema
        )
        response = await self._client._prepared_request(request)
        return Mailbox.model_validate(response)

    async def delete(self, email: str) -> None:
        request = self.prepare_delete(email)
        await self._client._prepared_request(request)
        return None


class TestMailboxesMixin:
    def prepare_forward(
        self,
        email: str,
        document: Path | str | IOBase | HttpUrl | MIMEData,
        verbose: bool = True,
    ) -> PreparedRequest:
        mime_document = prepare_mime_document(document)
        return PreparedRequest(method="POST", url=f"/v1/processors/automations/mailboxes/tests/forward/{email}", data={"document": mime_document.model_dump()})

    def print_forward_verbose(self, email_data: EmailData) -> None:
        print(f"\nTEST EMAIL FORWARDING RESULTS:")
        print(f"\n#########################")
        print(f"Email ID: {email_data.id}")
        print(f"Subject: {email_data.subject}")
        print(f"From: {email_data.sender}")
        print(f"To: {', '.join(str(r) for r in email_data.recipients_to)}")
        if email_data.recipients_cc:
            print(f"CC: {', '.join(str(r) for r in email_data.recipients_cc)}")
        print(f"Sent at: {email_data.sent_at}")
        print(f"Attachments: {len(email_data.attachments)}")
        if email_data.body_plain:
            print("\nBody Preview:")
            print(email_data.body_plain[:500] + "..." if len(email_data.body_plain) > 500 else email_data.body_plain)

class TestMailboxes(SyncAPIResource, TestMailboxesMixin):
    def forward(
        self,
        email: str,
        document: Path | str | IOBase | HttpUrl | MIMEData,
        verbose: bool = True,
    ) -> EmailData:
        """Mock endpoint that simulates the complete email forwarding process with sample data.

        Args:
            email: Email address of the mailbox to mock

        Returns:
            DocumentExtractResponse: The simulated extraction response
        """
        request = self.prepare_forward(email, document, verbose)
        response = self._client._prepared_request(request)

        email_data = EmailData.model_validate(response)

        if verbose:
            self.print_forward_verbose(email_data)
        return email_data

class AsyncTestMailboxes(AsyncAPIResource, TestMailboxesMixin):
    async def forward(
        self,
        email: str,
        document: Path | str | IOBase | HttpUrl | MIMEData,
        verbose: bool = True,
    ) -> EmailData:
        request = self.prepare_forward(email, document, verbose)
        response = await self._client._prepared_request(request)
        email_data = EmailData.model_validate(response)
        if verbose:
            self.print_forward_verbose(email_data)
        return email_data



import mimetypes
from typing import BinaryIO, Literal, Tuple

from pydantic import BaseModel, ConfigDict, Field, HttpUrl, field_serializer


class DBFile(BaseModel):
    """Represents the core file object in your new spec."""

    object: Literal["file"] = "file"
    id: str = Field(..., description="The unique identifier of the file. It is generated from the content 'file_{sha256(content)}'")
    filename: str = Field(..., description="The name of the file")

    @property
    def mime_type(self) -> str:
        return mimetypes.guess_type(self.filename)[0] or "application/octet-stream"

    @property
    def extension(self) -> str:
        return self.filename.split(".")[-1].lower()

    model_config = ConfigDict(arbitrary_types_allowed=True)


FileData = Tuple[str, BinaryIO, str]
FileTuple = Tuple[str, FileData]


class FileLink(BaseModel):
    download_url: HttpUrl = Field(description="The signed URL to download the file")
    expires_in: str = Field(description="The expiration time of the signed URL")
    filename: str = Field(description="The name of the file")

    @field_serializer('download_url')
    def url2str(self, val: HttpUrl) -> str:
        return str(val)



import datetime
from pathlib import Path
from typing import Any, Dict, Literal, Optional

import httpx
from openai.types.chat.chat_completion_reasoning_effort import ChatCompletionReasoningEffort
from pydantic import HttpUrl

from ...._resource import AsyncAPIResource, SyncAPIResource
from ...._utils.ai_models import assert_valid_model_extraction
from ....types.automations.endpoints import Endpoint, ListEndpoints, UpdateEndpointRequest
from ....types.logs import ExternalRequestLog

# from ...types.documents.extractions import DocumentExtractResponse
from ....types.mime import BaseMIMEData, MIMEData
from ....types.modalities import Modality
from ....types.standards import PreparedRequest


class EndpointsMixin:
    def prepare_create(
        self,
        name: str,
        webhook_url: HttpUrl,
        json_schema: Dict[str, Any],
        webhook_headers: Optional[Dict[str, str]] = None,
        # DocumentExtraction Config
        image_resolution_dpi: Optional[int] = None,
        browser_canvas: Optional[Literal['A3', 'A4', 'A5']] = None,
        modality: Modality = "native",
        model: str = "gpt-4o-mini",
        temperature: float = 0,
        reasoning_effort: ChatCompletionReasoningEffort = "medium",
    ) -> PreparedRequest:
        assert_valid_model_extraction(model)

        data = {
            "name": name,
            "webhook_url": webhook_url,
            "webhook_headers": webhook_headers or {},
            "json_schema": json_schema,
            "image_resolution_dpi": image_resolution_dpi,
            "browser_canvas": browser_canvas,
            "modality": modality,
            "model": model,
            "temperature": temperature,
            "reasoning_effort": reasoning_effort,
        }

        request = Endpoint.model_validate(data)
        return PreparedRequest(method="POST", url="/v1/deployments/endpoints", data=request.model_dump(mode='json'))

    def prepare_list(
        self,
        before: Optional[str] = None,
        after: Optional[str] = None,
        limit: Optional[int] = 10,
        order: Optional[Literal["asc", "desc"]] = "desc",
        # Filtering parameters
        name: Optional[str] = None,
        webhook_url: Optional[str] = None,
    ) -> PreparedRequest:
        params = {
            "before": before,
            "after": after,
            "limit": limit,
            "order": order,
            "name": name,
            "webhook_url": webhook_url,
        }
        # Remove None values
        params = {k: v for k, v in params.items() if v is not None}

        return PreparedRequest(method="GET", url="/v1/deployments/endpoints", params=params)

    def prepare_get(self, endpoint_id: str) -> PreparedRequest:
        """Get a specific endpoint configuration.

        Args:
            endpoint_id: ID of the endpoint

        Returns:
            Endpoint: The endpoint configuration
        """
        return PreparedRequest(method="GET", url=f"/v1/processors/automations/endpoints/{endpoint_id}")

    def prepare_update(
        self,
        endpoint_id: str,
        name: Optional[str] = None,
        webhook_url: Optional[HttpUrl] = None,
        webhook_headers: Optional[Dict[str, str]] = None,
        json_schema: Optional[Dict[str, Any]] = None,
        image_resolution_dpi: Optional[int] = None,
        browser_canvas: Optional[Literal['A3', 'A4', 'A5']] = None,
        modality: Optional[Modality] = None,
        model: Optional[str] = None,
        temperature: Optional[float] = None,
        reasoning_effort: Optional[ChatCompletionReasoningEffort] = None,
    ) -> PreparedRequest:
        data: dict[str, Any] = {}

        if endpoint_id is not None:
            data["id"] = endpoint_id
        if name is not None:
            data["name"] = name
        if webhook_url is not None:
            data["webhook_url"] = webhook_url
        if webhook_headers is not None:
            data["webhook_headers"] = webhook_headers
        if json_schema is not None:
            data["json_schema"] = json_schema
        if image_resolution_dpi is not None:
            data["image_resolution_dpi"] = image_resolution_dpi
        if browser_canvas is not None:
            data["browser_canvas"] = browser_canvas
        if modality is not None:
            data["modality"] = modality
        if model is not None:
            assert_valid_model_extraction(model)
            data["model"] = model
        if temperature is not None:
            data["temperature"] = temperature
        if reasoning_effort is not None:
            data["reasoning_effort"] = reasoning_effort
        request = UpdateEndpointRequest.model_validate(data)
        return PreparedRequest(method="PUT", url=f"/v1/processors/automations/endpoints/{endpoint_id}", data=request.model_dump(mode='json'))

    def prepare_delete(self, endpoint_id: str) -> PreparedRequest:
        return PreparedRequest(method="DELETE", url=f"/v1/processors/automations/endpoints/{endpoint_id}")


class Endpoints(SyncAPIResource, EndpointsMixin):
    """Endpoints API wrapper for managing endpoint configurations"""

    def create(
        self,
        name: str,
        webhook_url: HttpUrl,
        json_schema: Dict[str, Any],
        webhook_headers: Optional[Dict[str, str]] = None,
        # DocumentExtraction Config
        image_resolution_dpi: Optional[int] = None,
        browser_canvas: Optional[Literal['A3', 'A4', 'A5']] = None,
        modality: Modality = "native",
        model: str = "gpt-4o-mini",
        temperature: float = 0,
        reasoning_effort: ChatCompletionReasoningEffort = "medium",
    ) -> Endpoint:
        """Create a new endpoint configuration.

        Args:
            name: Name of the endpoint
            webhook_url: Webhook endpoint URL
            json_schema: JSON schema for the endpoint
            webhook_headers: Optional HTTP headers for webhook requests
            image_resolution_dpi: Optional image resolution DPI
            browser_canvas: Optional browser canvas size
            modality: Processing modality (currently only "native" supported)
            model: AI model to use for processing
            temperature: Model temperature setting
            reasoning_effort: The effort level for the model to reason about the input data.
        Returns:
            Endpoint: The created endpoint configuration
        """
        request = self.prepare_create(name, webhook_url, json_schema, webhook_headers, image_resolution_dpi, browser_canvas, modality, model, temperature, reasoning_effort)
        response = self._client._prepared_request(request)
        print(f"Endpoint ID: {response['id']}. Endpoint available at https://www.uiform.com/dashboard/processors/{response['id']}")
        return Endpoint.model_validate(response)

    def list(
        self,
        before: Optional[str] = None,
        after: Optional[str] = None,
        limit: Optional[int] = 10,
        order: Optional[Literal["asc", "desc"]] = "desc",
        name: Optional[str] = None,
        webhook_url: Optional[str] = None,
    ) -> ListEndpoints:
        """List endpoint configurations with pagination support.

        Args:
            before: Optional cursor for pagination before a specific endpoint ID
            after: Optional cursor for pagination after a specific endpoint ID
            limit: Optional limit on number of results (max 100)
            order: Optional sort order ("asc" or "desc")
            name: Optional filter by endpoint name
            webhook_url: Optional filter by webhook URL

        Returns:
            ListEndpoints: Paginated list of endpoint configurations with metadata
        """
        request = self.prepare_list(before, after, limit, order, name, webhook_url)
        response = self._client._prepared_request(request)
        return ListEndpoints.model_validate(response)

    def get(self, endpoint_id: str) -> Endpoint:
        """Get a specific endpoint configuration.

        Args:
            endpoint_id: ID of the endpoint

        Returns:
            Endpoint: The endpoint configuration
        """
        request = self.prepare_get(endpoint_id)
        response = self._client._prepared_request(request)
        return Endpoint.model_validate(response)

    def update(
        self,
        endpoint_id: str,
        name: Optional[str] = None,
        webhook_url: Optional[HttpUrl] = None,
        webhook_headers: Optional[Dict[str, str]] = None,
        json_schema: Optional[Dict[str, Any]] = None,
        image_resolution_dpi: Optional[int] = None,
        browser_canvas: Optional[Literal['A3', 'A4', 'A5']] = None,
        modality: Optional[Modality] = None,
        model: Optional[str] = None,
        temperature: Optional[float] = None,
        reasoning_effort: Optional[ChatCompletionReasoningEffort] = None,
    ) -> Endpoint:
        """Update an endpoint configuration.

        Args:
            id: ID of the endpoint to update
            name: New name for the endpoint
            webhook_url: New webhook URL
            webhook_headers: New webhook headers
            json_schema: New JSON schema for the endpoint
            image_resolution_dpi: New image resolution DPI
            browser_canvas: New browser canvas size
            modality: New processing modality
            model: New AI model
            temperature: New temperature setting
            reasoning_effort: The effort level for the model to reason about the input data.
        Returns:
            Endpoint: The updated endpoint configuration
        """
        request = self.prepare_update(endpoint_id, name, webhook_url, webhook_headers, json_schema, image_resolution_dpi, browser_canvas, modality, model, temperature, reasoning_effort)
        response = self._client._prepared_request(request)
        return Endpoint.model_validate(response)

    def delete(self, endpoint_id: str) -> None:
        """Delete an endpoint configuration.

        Args:
            id: ID of the endpoint to delete
        """
        request = self.prepare_delete(endpoint_id)
        self._client._prepared_request(request)
        print(f"Endpoint Deleted. ID: {endpoint_id}")


class AsyncEndpoints(AsyncAPIResource, EndpointsMixin):
    """Async Endpoints API wrapper for managing endpoint configurations"""

    async def create(
        self,
        name: str,
        webhook_url: HttpUrl,
        json_schema: Dict[str, Any],
        webhook_headers: Optional[Dict[str, str]] = None,
        # DocumentExtraction Config
        image_resolution_dpi: Optional[int] = None,
        browser_canvas: Optional[Literal['A3', 'A4', 'A5']] = None,
        modality: Modality = "native",
        model: str = "gpt-4o-mini",
        temperature: float = 0,
        reasoning_effort: ChatCompletionReasoningEffort = "medium",
    ) -> Endpoint:
        request = self.prepare_create(name, webhook_url, json_schema, webhook_headers, image_resolution_dpi, browser_canvas, modality, model, temperature, reasoning_effort)
        response = await self._client._prepared_request(request)
        print(f"Endpoint ID: {response['id']}. Endpoint available at https://www.uiform.com/dashboard/processors/{response['id']}")
        
        return Endpoint.model_validate(response)

    async def list(
        self,
        before: Optional[str] = None,
        after: Optional[str] = None,
        limit: Optional[int] = 10,
        order: Optional[Literal["asc", "desc"]] = "desc",
        name: Optional[str] = None,
        webhook_url: Optional[str] = None,
    ) -> ListEndpoints:
        request = self.prepare_list(before, after, limit, order, name, webhook_url)
        response = await self._client._prepared_request(request)
        return ListEndpoints.model_validate(response)

    async def get(self, endpoint_id: str) -> Endpoint:
        request = self.prepare_get(endpoint_id)
        response = await self._client._prepared_request(request)
        return Endpoint.model_validate(response)

    async def update(
        self,
        endpoint_id: str,
        name: Optional[str] = None,
        webhook_url: Optional[HttpUrl] = None,
        webhook_headers: Optional[Dict[str, str]] = None,
        json_schema: Optional[Dict[str, Any]] = None,
        image_resolution_dpi: Optional[int] = None,
        browser_canvas: Optional[Literal['A3', 'A4', 'A5']] = None,
        modality: Optional[Modality] = None,
        model: Optional[str] = None,
        temperature: Optional[float] = None,
        reasoning_effort: Optional[ChatCompletionReasoningEffort] = None,
    ) -> Endpoint:
        request = self.prepare_update(endpoint_id, name, webhook_url, webhook_headers, json_schema, image_resolution_dpi, browser_canvas, modality, model, temperature, reasoning_effort)
        response = await self._client._prepared_request(request)
        return Endpoint.model_validate(response)

    async def delete(self, endpoint_id: str) -> None:
        request = self.prepare_delete(endpoint_id)
        await self._client._prepared_request(request)



import os
import re
from typing import ClassVar, List, Literal, Optional

import nanoid  # type: ignore
from pydantic import BaseModel, EmailStr, Field, field_validator

from ..pagination import ListMetadata

domain_pattern = re.compile(r"^(?:[a-zA-Z0-9-]+\.)+[a-zA-Z]{2,}$")

from openai.types.chat.chat_completion_reasoning_effort import ChatCompletionReasoningEffort

from ..logs import AutomationConfig, UpdateAutomationRequest


class Mailbox(AutomationConfig):
    EMAIL_PATTERN: ClassVar[str] = f".*@{os.getenv('EMAIL_DOMAIN', 'mailbox.uiform.com')}$"
    object: Literal['automation.mailbox'] = "automation.mailbox"
    id: str = Field(default_factory=lambda: "mb_" + nanoid.generate(), description="Unique identifier for the mailbox")

    # Email Specific config
    email: str = Field(..., pattern=EMAIL_PATTERN)
    authorized_domains: list[str] = Field(default_factory=list, description="List of authorized domains to receive the emails from")
    authorized_emails: List[EmailStr] = Field(default_factory=list, description="List of emails to access the link")

    # Normalize email fields (case-insensitive)
    @field_validator("email", mode="before")
    def normalize_email(cls, value: str) -> str:
        return value.strip().lower()

    @field_validator("authorized_emails", mode="before")
    def normalize_authorized_emails(cls, emails: List[str]) -> List[str]:
        return [email.strip().lower() for email in emails]

    @field_validator('authorized_domains', mode='before')
    def validate_domain(cls, list_domains: list[str]) -> list[str]:
        for domain in list_domains:
            if not domain_pattern.match(domain):
                raise ValueError(f"Invalid domain: {domain}")
        return list_domains


class ListMailboxes(BaseModel):
    data: list[Mailbox]
    list_metadata: ListMetadata


# Inherits from the methods of UpdateAutomationRequest
class UpdateMailboxRequest(UpdateAutomationRequest):

    # ------------------------------
    # Email Specific config
    # ------------------------------
    authorized_domains: Optional[list[str]] = None
    authorized_emails: Optional[List[EmailStr]] = None

    @field_validator("authorized_emails", mode="before")
    def normalize_authorized_emails(cls, emails: Optional[List[str]]) -> Optional[List[str]]:
        return [email.strip().lower() for email in emails] if emails else None



from setuptools import find_packages, setup  # type: ignore

# Read requirements.txt and use it for the install_requires parameter
with open('requirements.txt') as f:
    requirements_list = f.read().splitlines()

setup(
    name='retab',
    version='0.0.36',
    author='UiForm',
    author_email='contact@uiform.com',
    description='UiForm official python library',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/UiForm/uiform',
    project_urls={'Team website': 'https://uiform.com'},
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX :: Linux',
        'Operating System :: MacOS',
        'Intended Audience :: Science/Research',
    ],
    packages=find_packages(),
    python_requires='>=3.6',
    install_requires=requirements_list,
    include_package_data=True,
)



import copy
import datetime
import json
from typing import Any, Dict, List, Literal, Optional

import nanoid  # type: ignore
from openai import OpenAI
from openai.types.chat.chat_completion_reasoning_effort import ChatCompletionReasoningEffort
from pydantic import BaseModel, EmailStr, Field, HttpUrl, computed_field, field_serializer
from pydantic_core import Url

from .._utils.json_schema import clean_schema, compute_schema_data_id
from .._utils.mime import generate_blake2b_hash_from_string
from .._utils.usage.usage import compute_cost_from_model, compute_cost_from_model_with_breakdown, CostBreakdown
from .ai_models import Amount
from .documents.extractions import UiParsedChatCompletion
from .mime import BaseMIMEData
from .modalities import Modality
from .pagination import ListMetadata


class ProcessorConfig(BaseModel):
    object: str = Field(default="processor", description="Type of the object")
    id: str = Field(default_factory=lambda: "proc_" + nanoid.generate(), description="Unique identifier for the processor")
    updated_at: datetime.datetime = Field(default_factory=lambda: datetime.datetime.now(datetime.timezone.utc))
    name: str = Field(..., description="Name of the processor")

    modality: Modality
    image_resolution_dpi: int = Field(default=96, description="Resolution of the image sent to the LLM")
    browser_canvas: Literal['A3', 'A4', 'A5'] = Field(default='A4', description="Sets the size of the browser canvas for rendering documents in browser-based processing. Choose a size that matches the document type.")

    # New attributes
    model: str = Field(..., description="Model used for chat completion")
    json_schema: dict[str, Any] = Field(..., description="JSON schema format used to validate the output data.")
    temperature: float = Field(default=0.0, description="Temperature for sampling. If not provided, the default temperature for the model will be used.", examples=[0.0])
    reasoning_effort: ChatCompletionReasoningEffort = Field(
        default="medium", description="The effort level for the model to reason about the input data. If not provided, the default reasoning effort for the model will be used."
    )
    n_consensus: int = Field(default=1, description="Number of consensus required to validate the data")

    @computed_field  # type: ignore
    @property
    def schema_data_id(self) -> str:
        """Returns the SHA1 hash of the schema data, ignoring all prompt/description/default fields.

        Returns:
            str: A SHA1 hash string representing the schema data version.
        """
        return compute_schema_data_id(self.json_schema)

    # This is a computed field, it is exposed when serializing the object
    @computed_field  # type: ignore
    @property
    def schema_id(self) -> str:
        """Returns the SHA1 hash of the complete schema.

        Returns:
            str: A SHA1 hash string representing the complete schema version.
        """
        return "sch_id_" + generate_blake2b_hash_from_string(json.dumps(self.json_schema, sort_keys=True).strip())


class AutomationConfig(BaseModel):
    object: str = Field(default="automation", description="Type of the object")
    id: str = Field(default_factory=lambda: "auto_" + nanoid.generate(), description="Unique identifier for the automation")
    name: str = Field(..., description="Name of the automation")
    processor_id: str = Field(..., description="ID of the processor to use for the automation")
    updated_at: datetime.datetime = Field(default_factory=lambda: datetime.datetime.now(datetime.timezone.utc), description="Timestamp of the last update")

    default_language: str = Field(default="en", description="Default language for the automation")

    # HTTP Config
    webhook_url: HttpUrl = Field(..., description="Url of the webhook to send the data to")
    webhook_headers: Dict[str, str] = Field(default_factory=dict, description="Headers to send with the request")

    need_validation: bool = Field(default=False, description="If the automation needs to be validated before running")

    @field_serializer('webhook_url')
    def url2str(self, val: HttpUrl) -> str:
        return str(val)


class UpdateProcessorRequest(BaseModel):
    # ------------------------------
    # Processor Parameters
    # ------------------------------
    name: Optional[str] = None
    modality: Optional[Modality] = None
    image_resolution_dpi: Optional[int] = None
    browser_canvas: Optional[Literal['A3', 'A4', 'A5']] = None
    model: Optional[str] = None
    json_schema: Optional[Dict] = None
    temperature: Optional[float] = None
    reasoning_effort: Optional[ChatCompletionReasoningEffort] = None
    n_consensus: Optional[int] = None

    @computed_field  # type: ignore
    @property
    def schema_data_id(self) -> Optional[str]:
        """Returns the SHA1 hash of the schema data, ignoring all prompt/description/default fields.

        Returns:
            str: A SHA1 hash string representing the schema data version.
        """
        if self.json_schema is None:
            return None
        return compute_schema_data_id(self.json_schema)

    @computed_field  # type: ignore
    @property
    def schema_id(self) -> Optional[str]:
        """Returns the SHA1 hash of the complete schema.

        Returns:
            str: A SHA1 hash string representing the complete schema version.
        """
        if self.json_schema is None:
            return None
        return "sch_id_" + generate_blake2b_hash_from_string(json.dumps(self.json_schema, sort_keys=True).strip())


class UpdateAutomationRequest(BaseModel):
    name: Optional[str] = None
    processor_id: Optional[str] = None

    default_language: Optional[str] = None

    webhook_url: Optional[HttpUrl] = None
    webhook_headers: Optional[Dict[str, str]] = None
    
    need_validation: Optional[bool] = None


    @field_serializer('webhook_url')
    def url2str(self, val: HttpUrl | None) -> str | None:
        if isinstance(val, HttpUrl):
            return str(val)
        return val

class OpenAIRequestConfig(BaseModel):
    object: Literal['openai_request'] = "openai_request"
    id: str = Field(default_factory=lambda: "openai_req_" + nanoid.generate(), description="Unique identifier for the openai request")
    model: str
    json_schema: dict[str, Any]
    reasoning_effort: Optional[ChatCompletionReasoningEffort] = None


# ------------------------------
# ------------------------------
# ------------------------------

# from .automations.mailboxes import Mailbox
# from .automations.links import Link
# from .automations.cron import ScrappingConfig
# from .automations.outlook import Outlook

# class OpenAILog(BaseModel):
#    request_config: OpenAIRequestConfig
#    completion: ChatCompletion


class ExternalRequestLog(BaseModel):
    webhook_url: Optional[HttpUrl]
    request_body: dict[str, Any]
    request_headers: dict[str, str]
    request_at: datetime.datetime

    response_body: dict[str, Any]
    response_headers: dict[str, str]
    response_at: datetime.datetime

    status_code: int
    error: Optional[str] = None
    duration_ms: float

    @field_serializer('webhook_url')
    def url2str(self, val: HttpUrl | None) -> str | None:
        if isinstance(val, HttpUrl):
            return str(val)
        return val


from openai.types.chat import completion_create_params
from openai.types.chat.chat_completion import ChatCompletion


class LogCompletionRequest(BaseModel):
    json_schema: dict[str, Any]
    completion: ChatCompletion


class AutomationLog(BaseModel):
    object: Literal['automation_log'] = "automation_log"
    id: str = Field(default_factory=lambda: "log_auto_" + nanoid.generate(), description="Unique identifier for the automation log")
    user_email: Optional[EmailStr]  # When the user is logged or when he forwards an email
    organization_id: str
    created_at: datetime.datetime = Field(default_factory=lambda: datetime.datetime.now(datetime.timezone.utc))
    automation_snapshot: AutomationConfig
    completion: UiParsedChatCompletion | ChatCompletion
    file_metadata: Optional[BaseMIMEData]
    external_request_log: Optional[ExternalRequestLog]
    extraction_id: Optional[str] = Field(default=None, description="ID of the extraction")

    @computed_field  # type: ignore
    @property
    def api_cost(self) -> Optional[Amount]:
        if self.completion and self.completion.usage:
            try:
                cost = compute_cost_from_model(self.completion.model, self.completion.usage)
                return cost
            except Exception as e:
                print(f"Error computing cost: {e}")
                return None
        return None
    
    @computed_field  # type: ignore
    @property
    def cost_breakdown(self) -> Optional[CostBreakdown]:
        if self.completion and self.completion.usage:
            try:
                cost = compute_cost_from_model_with_breakdown(self.completion.model, self.completion.usage)
                return cost
            except Exception as e:
                print(f"Error computing cost: {e}")
                return None
        return None


class ListLogs(BaseModel):
    data: List[AutomationLog]
    list_metadata: ListMetadata
